from dataclasses import dataclass, InitVar
from datetime import datetime
from typing import Optional

from vectice.models import model_version
from .model_output import ModelOutput


@dataclass
class __Output:
    createdDate: datetime
    updatedDate: datetime
    id: int
    versionNumber: int
    name: str
    description: Optional[str]
    uri: Optional[str]
    authorId: int
    deletedDate: Optional[datetime]
    modelId: int
    isStarred: str
    originId: Optional[int]
    model: InitVar[dict]

    def __post_init__(self, model: dict):
        self.model = ModelOutput(**model)


@dataclass
class ModelVersionOutput(model_version.ModelVersion, __Output):
    pass



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, TypeVar

from .job_type import JobType

JobAndChildTypes = TypeVar("JobAndChildTypes", bound="Job")


@dataclass
class Job:
    name: str
    """"""
    type: Optional[str] = JobType.OTHER
    """
    see possible values in py:class:: JobType
    """
    description: Optional[str] = None
    """"""

    def with_description(self: JobAndChildTypes, description: str) -> JobAndChildTypes:
        """"""
        self.description = description
        return self

    def with_type(self: JobAndChildTypes, job_type: str) -> JobAndChildTypes:
        """"""
        self.type = job_type
        return self



from __future__ import annotations

import logging
import os
import time

from abc import abstractmethod, ABC
from typing import List, Optional, Any, Dict, Sequence, Tuple, Union

from vectice.api import Client, Page
from vectice.api.json_object import JsonObject
from vectice.api.output import JobOutput, JobRunOutput, PagedResponse
from vectice.api.output.attachments_output import AttachmentOutput
from vectice.api.output.connection_output import ConnectionOutput
from vectice.api.output.dataset_output import DatasetOutput
from vectice.api.output.dataset_version_output import DatasetVersionOutput
from vectice.api.output.model_output import ModelOutput
from vectice.api.output.model_version_output import ModelVersionOutput
from vectice.entity.model import ModelType
from vectice.models import (
    Artifact,
    ArtifactType,
    RunnableJob,
    Job,
    JobRun,
    JobRunStatus,
    DatasetVersionArtifact,
    Artifacts,
    ModelVersion,
    ModelVersionArtifact,
    CodeVersionArtifact,
    DataResource,
    CreateDataset,
    Attachments,
    UpdateModelVersion,
    DatasetMetadataArtifact,
)
from vectice.utils.artifact_api_utils import get_artifact_id, get_artifact_version_id, update_artifact_properties


# TODO implement TypeAlias when we use python 3.10 e.g list_model_version_dataframe -> DataFrame
# DataFrame: TypeAlias = "DataFrame"


class AbstractAdapter(ABC):
    @property
    @abstractmethod
    def active_runs(self) -> Dict[int, ActiveRun]:
        pass

    @abstractmethod
    def create_run(self, name: str) -> RunnableJob:
        pass

    @abstractmethod
    def end_run(
        self, run: ActiveRun, outputs: Optional[List[Artifact]] = None, status: str = JobRunStatus.COMPLETED
    ) -> Optional[int]:
        pass

    @abstractmethod
    def start_run(self, run: RunnableJob, inputs: Optional[List[Artifact]] = None) -> ActiveRun:
        pass

    @abstractmethod
    def save_job_and_associated_runs(self, name: str) -> None:
        pass

    @abstractmethod
    def save_run(
        self,
        run: Any,
        inputs: Optional[List[Artifact]] = None,
        outputs: Optional[List[Artifact]] = None,
    ) -> Optional[int]:
        pass

    @abstractmethod
    def run_failed(self, run: Optional[ActiveRun] = None):
        pass


class ActiveRun:
    """Wrapper around dict response to enable using Python ``with`` syntax."""

    _outputs: Optional[List[Artifact]]
    _adapter: AbstractAdapter
    _job: JsonObject
    _run: JsonObject
    _inputs: JsonObject

    def __init__(self, job: JsonObject, run: JsonObject, inputs: JsonObject, adapter: AbstractAdapter):
        self._adapter = adapter
        self._job = job
        self._run = run
        self._inputs = inputs
        self._outputs = None
        self._logger = logging.getLogger(self.__class__.__name__)

    def __enter__(self) -> ActiveRun:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        status = JobRunStatus.COMPLETED if exc_type is None else JobRunStatus.FAILED
        try:
            self._adapter.end_run(self, status=status)
        except Exception as e:
            self._adapter.run_failed(self)
            self._logger.warning(f"Run failed because {e}")
        finally:
            return exc_type is None

    @property
    def outputs(self) -> Optional[List[Artifact]]:
        return self._outputs

    @property
    def run(self) -> JsonObject:
        return self._run

    @property
    def job(self) -> JsonObject:
        return self._job

    def add_output(self, output: Artifact):
        if self._outputs is None:
            self._outputs = []
        self._outputs.append(output)

    def add_outputs(self, outputs: List[Artifact]):
        if len(outputs) > 0:
            if self._outputs is None:
                self._outputs = []
            self._outputs.extend(outputs)


class Adapter(AbstractAdapter):
    def __init__(
        self,
        project_token: str,
        auto_connect: bool = True,
        autocode: bool = False,
        allow_self_signed_certificate: bool = True,
    ):
        self._client = Client(project_token, auto_connect, allow_self_signed_certificate)
        self._active_runs: Dict[int, ActiveRun] = {}
        self._last_created_run: Optional[RunnableJob] = None
        self._last_started_run: Optional[ActiveRun] = None
        self._logger = logging.getLogger(self.__class__.__name__)
        self.autocode = autocode
        try:
            self._client.auth_project_token(project_token)
        except Exception:
            raise ValueError(
                "The entered token is invalid. Please check the entered value. You may find and copy a Project Token directly under the Settings menu, on your Project"
            )
        self._vectice_path = self._client.vectice_path

    @property
    def active_runs(self) -> Dict[int, ActiveRun]:
        return self._active_runs

    @property
    def vectice_path(self) -> Optional[str]:
        """
        Returns the project and workspace being used.

        :return: A string of the Vectice path
        """
        self._logger.info(self._vectice_path)
        return self._vectice_path

    def get_current_runnable_job(self, run: Optional[RunnableJob] = None) -> RunnableJob:
        if run is not None:
            result: RunnableJob = run
        else:
            if self._last_created_run is None:
                raise RuntimeError("A job context must have been created.")
            else:
                result = self._last_created_run
        return result

    def start_run(self, run: Optional[RunnableJob] = None, inputs: Optional[List[Artifact]] = None) -> ActiveRun:
        """
        Start the run created before by calling :func:`~Vectice.create_run` function

        :param run: The runnable job to start
        :param inputs: A list of artifacts used as inputs by this run.
        :return: A reference to a run in progress
        """

        run = self.get_current_runnable_job(run)
        code_artifact_is_present = False
        artifact_name, attachments = None, None
        if inputs is not None:
            for an_input in inputs:
                if an_input is not None:
                    an_input.jobArtifactType = "INPUT"
                    code_artifact_is_present = code_artifact_is_present or an_input.artifactType == ArtifactType.CODE
                    artifact_name, attachments = self._parse_attachments(an_input)
        if not code_artifact_is_present and self.autocode:
            if inputs is None:
                inputs = []
            artifact = CodeVersionArtifact.create(".")
            if artifact is not None:
                inputs.append(artifact)

        response = self._client.start_run(run, inputs)
        if attachments and attachments.files:
            self._save_attachments(artifact_name, attachments)
        active_run = ActiveRun(response["job"], response["jobRun"], response["jobArtifacts"], self)
        self._active_runs[active_run.run["id"]] = active_run
        self._last_started_run = active_run
        return active_run

    def _get_current_active_run(self, run: Optional[ActiveRun] = None) -> ActiveRun:
        if run is not None:
            result: ActiveRun = run
        else:
            if self._last_started_run is None:
                raise RuntimeError("A job context must have been created.")
            else:
                result = self._last_started_run
        return result

    def end_run(
        self,
        run: Optional[ActiveRun] = None,
        outputs: Optional[List[Artifact]] = None,
        status: str = JobRunStatus.COMPLETED,
    ) -> Optional[int]:
        """
        End the current (last) active run started by :func:`~Vectice.start_run`.
        To end a specific run, use :func:`~Vectice.stop_run` instead.

        :return: Identifier of the run in Vectice if successfully saved
        """
        run = self._get_current_active_run(run)
        artifact_name, attachments = None, None
        if outputs is not None:
            run.add_outputs(outputs)
        if run.outputs is not None:
            for an_output in run.outputs:
                if an_output is not None:
                    artifact_name, attachments = self._parse_attachments(an_output)
                    an_output.jobArtifactType = "OUTPUT"
        run.run["status"] = status
        self._client.stop_run(run.run, run.outputs)
        if attachments and attachments.files:
            self._save_attachments(artifact_name, attachments)
        if "id" in run.run:
            run_id: Optional[int] = int(run.run["id"])
        else:
            run_id = None
        del self._active_runs[run.run["id"]]
        return run_id

    # TODO Update to handle more artifactTypes
    def _save_attachments(self, artifact_name: Optional[str], attachments: Attachments):
        try:
            model_id = self.list_models(artifact_name).list[0].id
            version_id = self.list_model_versions(model_id).list[0].id
            self._client.create_attachments("modelversion", version_id, files=attachments.files)
        except Exception as e:
            self._logger.warning(f"Attachments failed to upload due to {e}")

    @staticmethod
    def _parse_attachments(an_input: Optional[Artifact] = None):
        artifact_name, attachments = None, None
        if isinstance(an_input, ModelVersionArtifact) and an_input.artifactType == ArtifactType.MODEL:
            attachments = an_input.attachments
            artifact_name = an_input.model.parentName
            # remove attachments from payload to avoid any unwanted behaviour
            an_input.attachments = Attachments(None, None)
        return artifact_name, attachments

    def __save_run(
        self,
        run: Optional[RunnableJob] = None,
        inputs: Optional[List[Artifact]] = None,
        outputs: Optional[List[Artifact]] = None,
    ) -> Optional[int]:
        if run is None:
            run = self._last_created_run
        active_run = self.start_run(run, inputs)
        return self.end_run(active_run, outputs)

    def save_job_and_associated_runs(self, name: str) -> None:
        raise RuntimeError("No implementation for this library")

    def save_run(
        self,
        run: Any,
        inputs: Optional[List[Artifact]] = None,
        outputs: Optional[List[Artifact]] = None,
    ) -> Optional[int]:
        """
        Save run with its associated inputs and outputs

        :param run: The run we want to save
        :param inputs: A list of inputs (artifacts) you are using in this run
        :param outputs: A list of outputs (artifacts) you are using in this run
        :return: Identifier of the run in Vectice if successfully saved
        """
        if isinstance(run, RunnableJob):
            return self.__save_run(run, inputs, outputs)
        else:
            raise RuntimeError("Incompatible object provided.")

    def create_run(self, job_name: str, job_type: Optional[str] = None, run_name: Optional[str] = None) -> RunnableJob:
        """
        Create an instance of a future run of a job.
        The run is not started.
        You need to start it by calling start_run

        :param job_type: The type of the job. see :class:`~vectice.models.JobType` for the list of accepted type.
        :param job_name: The name of the job that should run.
        :param run_name: The name of the run.
        :return: An instance of a non started run.
        """
        if job_name is None:
            raise RuntimeError("Job name must be set")
        self._last_created_run = RunnableJob(Job(job_name, job_type), JobRun(name=run_name))
        return self._last_created_run

    def run_failed(self, run: Optional[ActiveRun] = None):
        """
        Indicates that the run failed
        """
        self.__update_run_status(run, JobRunStatus.FAILED)

    def run_aborted(self, run: Optional[ActiveRun] = None):
        """
        Indicates that the run was aborted by the user
        """
        self.__update_run_status(run, JobRunStatus.ABORTED)

    def __update_run_status(self, active_run: Optional[ActiveRun], status: str):
        try:
            active_run = self._get_current_active_run(active_run)
            active_run.run["status"] = status
            self._client.update_run(active_run.job["id"], active_run.run["id"], active_run.run)
        except RuntimeError:
            logging.error("run failed to start.")

    @staticmethod
    def setup_api_token(token_file: str, api_endpoint: Optional[str] = None) -> None:
        """
        Parses the the API Token json file and sets the os environmental variable for the "VECTICE_API_TOKEN"
        for the user.

        :param token_file: The filepath to the json file containing the "VECTICE_API_TOKEN", found on the Vectice App
        :param api_endpoint: The api endpoint used to connect to the Vectice App
        :return: None
        """
        try:
            token = Artifacts.parse_api_token(token_file)
        except Exception as e:
            raise RuntimeError(f"File failed to parse: {e}")
        logging.info("Api token json file Parsed!")
        if api_endpoint:
            os.environ["VECTICE_API_ENDPOINT"] = api_endpoint
        Client.auth_api_token(api_token=token.key, api_endpoint=api_endpoint)
        os.environ["VECTICE_API_TOKEN"] = token.key
        logging.info(f"{token.name} parsed and configured!")

    def list_jobs(
        self, search: Optional[str] = None, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[JobOutput]:
        """
        List all jobs

        :param search: A text to filter jobs base on their name
        :param page_index: The index of the page we want
        :param page_size: The size of the page we want
        :return: A list of filtered jobs
        """
        return self._client.list_jobs(search, page_index, page_size)

    def list_runs(self, job_id: int, page_index=Page.index, page_size=Page.size) -> Sequence[JobRunOutput]:
        """
        List all run of a specific job

        :param job_id: The Vectice job identifier
        :param page_index: The index of the page we want
        :param page_size: The size of the page we want
        :return: A list of runs
        """
        return self._client.list_runs(job_id, page_index, page_size)

    def create_dataset(
        self,
        dataset_name: str,
        data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None,
    ) -> Optional[int]:
        """
        create a new dataset without a connection.

        :param dataset_name: the dataset name
        :param data_properties: the properties added to the dataset version, accepts either a tuple eg. (key, value) or a list of tuples with multiple properties eg. [(key, value), (key, value)]
        :return: identifier of the created dataset
        """
        dataset, dataset_id = None, None
        created_dataset = CreateDataset(name=dataset_name).add_properties(data_properties=data_properties).as_dict()
        dataset = self._client.create_dataset({"name": created_dataset["name"], "pattern": "*"})
        dataset_id = int(dataset["id"])
        if created_dataset.get("properties"):
            self._client.create_dataset_version(dataset_id=int(dataset["id"]), dataset_version=created_dataset)
        self._logger.info(f"Dataset: {dataset_name} has been successfully created.")
        return dataset_id

    def create_dataset_with_connection_id(
        self,
        connection_id: int,
        dataset_name: str,
        files: Optional[List[str]],
        folders: Optional[List[str]] = None,
        data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None,
    ) -> Optional[int]:
        """
        create a new dataset linked to the connection with associated items in it
        :param connection_id: the connection identifier
        :param dataset_name: the dataset name
        :param files: the list of files to be set in the dataset
        :param folders: the list of folders to be set in the dataset
        :param data_properties: the properties added to the dataset, accepts either a tuple eg. (key, value) or a list of tuples with multiple properties eg. [(key, value), (key, value)]
        :return: identifier of the created dataset
        """
        dataset_version_id, dataset, dataset_id = None, None, None
        data_resources = DataResource.create_resources(files, folders)
        created_dataset = (
            CreateDataset(name=dataset_name, connectionId=connection_id, files=files, folders=folders)
            .add_properties(data_properties=data_properties)
            .as_dict()
        )
        if data_resources:
            dataset = self._client.create_dataset(
                {
                    "name": created_dataset["name"],
                    "connectionId": created_dataset["connectionId"],
                    "dataResources": data_resources,
                }
            )
            time.sleep(1)
            dataset_id = int(dataset["id"])
            dataset_version_id = self._client.list_dataset_versions(dataset_id=int(dataset["id"])).list[0].id
        elif not data_resources:
            dataset = self._client.create_dataset(
                {"name": created_dataset["name"], "connectionId": created_dataset["connectionId"]}
            )
            dataset_id = int(dataset["id"])
            if created_dataset.get("properties"):
                self._client.create_dataset_version(dataset_id=int(dataset["id"]), dataset_version=created_dataset)
        if dataset and dataset_version_id and dataset_id and created_dataset.get("properties"):
            self._client.create_dataset_version_properties(
                dataset_id=dataset_id, dataset_version_id=dataset_version_id, properties=created_dataset["properties"]
            )
        self._logger.info(f"Dataset: {dataset_name} has been successfully created.")
        return dataset_id

    def create_dataset_with_connection_name(
        self,
        connection_name: str,
        dataset_name: str,
        files: Optional[List[str]],
        folders: Optional[List[str]] = None,
        data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None,
    ) -> Optional[int]:
        """
        create a new dataset linked to the connection with associated items in it
        :param connection_name: the connection name
        :param dataset_name: the dataset name
        :param files: the list of files to be set in the dataset
        :param folders: the list of folders to be set in the dataset
        :param data_properties: the properties added to the dataset, accepts either a tuple eg. (key, value) or a list of tuples with multiple properties eg. [(key, value), (key, value)]
        :return: identifier of the created dataset
        """
        dataset_version_id, dataset, dataset_id = None, None, None
        data_resources = DataResource.create_resources(files, folders)
        created_dataset = (
            CreateDataset(name=dataset_name, connectionName=connection_name, files=files, folders=folders)
            .add_properties(data_properties=data_properties)
            .as_dict()
        )
        if data_resources:
            dataset = self._client.create_dataset(
                {
                    "name": created_dataset["name"],
                    "connectionName": created_dataset["connectionName"],
                    "dataResources": data_resources,
                }
            )
            time.sleep(1)
            dataset_id = int(dataset["id"])
            dataset_version_id = self._client.list_dataset_versions(dataset_id=int(dataset["id"])).list[0].id
        elif not data_resources:
            dataset = self._client.create_dataset(
                {"name": created_dataset["name"], "connectionName": created_dataset["connectionName"]}
            )
            dataset_id = int(dataset["id"])
            if created_dataset.get("properties"):
                self._client.create_dataset_version(dataset_id=int(dataset["id"]), dataset_version=created_dataset)
        if dataset and dataset_version_id and dataset_id and created_dataset.get("properties"):
            self._client.create_dataset_version_properties(
                dataset_id=dataset_id, dataset_version_id=dataset_version_id, properties=created_dataset["properties"]
            )
        self._logger.info(f"Dataset: {dataset_name} has been successfully created.")
        return dataset_id

    @classmethod
    def create_dataset_version(cls, description: Optional[str] = None) -> DatasetVersionArtifact:
        """
        Create an artifact that contains a version of a dataset

        :param description: A description of the dataset version
        :return: A dataset version artifact
        """
        return Artifacts.create_dataset_version(description)

    def update_dataset(
        self, dataset_name: Optional[str] = None, dataset_id: Optional[int] = None, notes: Optional[str] = None
    ):
        """
        Update a dataset associated with the provided dataset name or dataset id.

        :param dataset_name: The name of the dataset to be updated
        :param dataset_id: The dataset id of the dataset to be updated
        :param notes: The dataset notes
        :return: None
        """
        dataset = None
        try:
            if dataset_name:
                dataset = self._client.list_datasets(dataset_name).list[0]
                dataset_id = dataset.id
                dataset.description = notes
            if dataset_id:
                dataset = [dataset for dataset in self._client.list_datasets().list if dataset.id == dataset_id][0]
                dataset_id = dataset.id
                dataset.description = notes
                self._client.update_dataset(dataset_id=dataset_id, dataset=dataset.__dict__)
                self._logger.info(f"Dataset '{dataset_name or dataset_id}' has been updated successfully")
        except Exception as e:
            raise ValueError(f"Dataset could not be updated due to {e}")

    def update_dataset_version(
        self,
        dataset_name: Optional[str] = None,
        dataset_id: Optional[int] = None,
        dataset_version_number: Optional[int] = None,
        dataset_version_id: Optional[int] = None,
        notes: Optional[str] = None,
        properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None,
    ):
        """
        Update a dataset version associated with the provided dataset name or dataset id and the dataset version number or dataset version id.

        properties can be as follows: [("key", "value"), ("key", "value")] or ("key", "value")

        :param dataset_name: The name of the dataset to be updated
        :param dataset_id: The dataset id of the dataset to be updated
        :param dataset_version_number: The dataset version number of the dataset to be updated
        :param dataset_version_id: The dataset version id of the dataset version to be updated
        :param notes: The dataset notes
        :param properties: The properties of the dataset to be updated or added
        :return: None
        """
        dataset = CreateDataset().add_properties(data_properties=properties).as_dict()
        try:
            if dataset_name:
                dataset_id = get_artifact_id(client=self._client, artifact_name=dataset_name)
            if dataset_version_number and dataset_id:
                dataset_version_id = get_artifact_version_id(
                    client=self._client, artifact_id=dataset_id, version_number=dataset_version_number
                )
            if dataset_id and dataset_version_id:
                if notes is not None:
                    self._client.update_dataset_version(
                        dataset_id=dataset_id,
                        dataset_version_id=dataset_version_id,
                        dataset_version={"description": notes},
                    )
                if dataset.get("properties"):
                    update_artifact_properties(self._client, dataset_id, dataset_version_id, dataset)
            self._logger.info(f"Dataset '{dataset_name or dataset_id}' has been updated successfully")
        except Exception as e:
            raise RuntimeError(f"Dataset version could not be updated due to {e}")

    def list_datasets(
        self, search: str = None, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[DatasetOutput]:
        """
        list datasets

        :param search:
        :param page_index:
        :param page_size:
        :return:
        """
        return self._client.list_datasets(search, page_index, page_size)

    def list_dataset_versions(
        self, dataset_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[DatasetVersionOutput]:
        """

        :param dataset_id:
        :param page_index:
        :param page_size:
        :return:
        """
        return self._client.list_dataset_versions(dataset_id, page_index, page_size)

    @classmethod
    def create_model_version(cls, description: Optional[str] = None) -> ModelVersionArtifact:
        """
        Create an artifact that contains a version of a model

        :param description: A description of the model version
        :return: A model version artifact
        """
        return Artifacts.create_model_version(description)

    def update_model(
        self,
        model_id: Optional[int] = None,
        parent_name: Optional[str] = None,
        model_type: Optional[Union[ModelType, str]] = None,
        description: Optional[str] = None,
    ) -> None:
        """
        Updates a model at the higher level and not the model version.

        :param model_id: The model's id
        :param parent_name: The model's name
        :param model_type: The model type e.g CLASSIFICATION, REGRESSION etc
        :param description: A description of the model
        :return: None
        """
        if not model_id:
            try:
                model_id = self._client.list_models(parent_name).list[0].id
            except Exception:
                raise ValueError(f"The model '{parent_name}' could not be found")
        if model_type:
            try:
                model_type = ModelVersionArtifact(ModelVersion(), Attachments()).with_type(model_type).model.type
            except Exception:
                raise ValueError(f"The model type of '{model_type}' isn't supported.")
        self._client.update_model(model_id, {"type": model_type, "description": description})
        model_identifier = parent_name if parent_name else model_id
        self._logger.info(f"Model: '{model_identifier}' has been updated")

    def update_model_version(
        self,
        model_name: Optional[str] = None,
        model_id: Optional[int] = None,
        model_version_id: Optional[int] = None,
        model_version_number: Optional[int] = None,
        notes: Optional[str] = None,
        algorithm_name: Optional[str] = None,
        properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None,
        files: Optional[List[str]] = None,
    ) -> None:
        """
        Updates a model version. The model_name or model_id are used to find the model. The model_version_id or model_version_number are used
        to find the model version to be updated.

        properties can be as follows: [("key", "value"), ("key", "value")] or ("key", "value")

        :param model_name: The name of the model to update
        :param model_id: The model id to update
        :param model_version_id: The model version's id
        :param model_version_number: The model version number to update
        :param notes: A description of the model
        :param algorithm_name: The model's algorithm name
        :param properties: The properties for the model version
        :param files: The files to be updated. If the filename exists, it will be updated
        :return: None
        """
        model = (
            UpdateModelVersion(
                algorithmName=algorithm_name,
                description=notes,
            )
            .add_properties(data_properties=properties)
            .as_dict()
        )
        try:
            if model_name and not model_id:
                model_id = get_artifact_id(client=self._client, artifact_name=model_name)
            if model_id and model_version_number and not model_version_id:
                model_version_id = get_artifact_version_id(
                    client=self._client, artifact_id=model_id, version_number=model_version_number
                )
            attachments = UpdateModelVersion.create_attachments(files)
            if attachments and attachments.artifact_type and model_version_id:
                self._logger.info("Updating model attachments...")
                self._client.update_attachments(
                    artifact_type=attachments.artifact_type, artifact_version=model_version_id, files=attachments.files
                )
                self._logger.info("Attachments updated successfully.")

            if model_id and model_version_id:
                if notes is not None:
                    self._client.update_model_version(
                        model_id=model_id, model_version_id=model_version_id, model_version={"description": notes}
                    )
                if algorithm_name is not None:
                    self._client.update_model_version(
                        model_id=model_id,
                        model_version_id=model_version_id,
                        model_version={"algorithmName": algorithm_name},
                    )
                if model.get("properties"):
                    update_artifact_properties(self._client, model_id, model_version_id, model)

            self._logger.info(f"Model Version '{model_name or model_id}' has been updated successfully")
        except Exception as e:
            raise RuntimeError(f"Model version could not be updated due to {e}")

    def list_models(self, search: str = None, page_index=Page.index, page_size=Page.size) -> PagedResponse[ModelOutput]:
        """

        :param search:
        :param page_index:
        :param page_size:
        :return:
        """
        return self._client.list_models(search, page_index, page_size)

    def list_model_versions(
        self, model_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[ModelVersionOutput]:
        """
        Lists model versions in a PagedResponse.

        :param model_id: The id of the model
        :param page_index: The page index e.g 1/20 would return page 1
        :param page_size: The amount of versions on a page
        :return: PagedResponse
        """
        return self._client.list_model_versions(model_id, page_index, page_size)

    def list_model_versions_dataframe(self, model_id: int, page_index=Page.index, page_size=Page.size):
        """
        Lists model versions in a pandas DataFrame and sorts by update date. Requires the pandas module to be installed, which can be done with
        ```pip install pandas``` or ```pip install vectice[pandas]```

        :param model_id: The id of the model
        :param page_index: The page index. For example 1/20 would return page 1
        :param page_size: The number of versions on a page
        :return: pd.DataFrame
        """
        import pandas as pd  # type: ignore

        models = self._client.list_model_versions(model_id, page_index, page_size)
        metric_dataframes = []
        property_dataframes = []
        for idx, model_version in enumerate(models.list, start=0):
            model_metrics = self._client.list_model_version_metrics(
                model_id, model_version.id, page_index, page_size
            ).list
            temp_df_metrics = pd.DataFrame(
                data={metrics.key: [metrics.value] for metrics in model_metrics}, index=[idx]
            )
            metric_dataframes += [temp_df_metrics]

        for idx, model_version in enumerate(models.list, start=0):
            models_properties = self._client.list_model_version_properties(
                model_id, model_version.id, page_index, page_size
            ).list
            temp_df_metrics = pd.DataFrame(
                data={metrics.key: [metrics.value] for metrics in models_properties}, index=[idx]
            )
            property_dataframes += [temp_df_metrics]

        metric_dataframes = pd.concat(metric_dataframes)
        property_dataframes = pd.concat(property_dataframes)
        merged_entities = pd.merge(metric_dataframes, property_dataframes, left_index=True, right_index=True)
        df_model = pd.DataFrame(
            data=models.list,
            columns=["createdDate", "name", "versionNumber", "status", "algorithmName", "isStarred"],
        )
        return pd.merge(df_model, merged_entities, left_index=True, right_index=True).sort_values(
            by=["versionNumber"], ascending=False
        )

    def list_model_version_attachments(self, model_version_id: int) -> PagedResponse[AttachmentOutput]:
        """
        List the attachments associated with a model version.

        :param model_version_id: The artifact version ID.
        :return: PagedResponse[AttachmentOutput]
        """
        return self._client.list_attachments("modelversion", model_version_id)

    def download_model_version_attachment(self, model_version_id: int, file_id: int):
        """
        Download an attachment associated with it's version. e.g model_version_id: 1120 and file_id: 21

        :param model_version_id: The model version ID.
        :param file_id: The ID of the file to be downloaded
        :return:
        """
        download = None
        try:
            download = self._client.get_attachment("modelversion", model_version_id, file_id)
        except Exception as e:
            raise RuntimeError(f"The file_id: {file_id} could not be downloaded due to: {e}")
        if download is None:
            raise ValueError(f"File_id: {file_id} failed to download. Ensure model version and file exists.")
        self._logger.info(f"File_id: {file_id} downloaded successfully.")
        return download

    def delete_model_version_attachment(self, model_version_id: int, file_id: int):
        """
        Delete an attachment associated with it's version. model_version_id: 1120 and file_id: 21

        :param model_version_id: The model version ID.
        :param file_id: The ID of the file to be deleted
        :return:
        """
        try:
            self._client.delete_attachment("modelversion", model_version_id, file_id)
        except Exception as e:
            raise RuntimeError(f"The file_id: {file_id} could not be deleted due to: {e}")
        self._logger.info(f"File_id: {file_id} has been successfully deleted.")

    def list_connections(
        self, search_name: Optional[str] = None, connection_type: Optional[str] = None
    ) -> PagedResponse[ConnectionOutput]:
        """
        List attachments associated to current workspace and project.

        :param search_name: Search the name of the connection
        :param connection_type: The connection type of the connections e.g 'GoogleStorage'
        :return:
        """
        return self._client.list_connections(search_name=search_name, connection_type=connection_type)

    @classmethod
    def create_code_version(cls, path: str = ".") -> Optional[CodeVersionArtifact]:
        """
        Create an artifact that contains a version of a code

        :param path: The path to the source code
        :return: A code version artifact
        """
        return Artifacts.create_code_version(path)

    @classmethod
    def create_code_version_with_github_uri(
        cls, uri: str, script_relative_path: Optional[str] = None, login_or_token=None, password=None, jwt=None
    ) -> Optional[CodeVersionArtifact]:
        """
        Create a code artifact based on the GitHub information relative to the given URI and relative path.

        Note: The URI given can include the branch you are working on. otherwise, the default repository branch will be used.

        sample :
            https://github.com/my-organization/my-repository (no branch given so using default branch)
            https://github.com/my-organization/my-repository/tree/my-current-branch (branch given is my-current-branch)

        To access private repositories, you need to authenticate with your credentials.
        see https://docs.github.com/en/github/authenticating-to-github/keeping-your-account-and-data-secure/about-authentication-to-github

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param login_or_token: A real login or a personal access token
        :param password: The password
        :param jwt: The OAuth2 access token
        :return: A CodeVersion or None if the GitHub repository was not found or is not accessible
        """
        return Artifacts.create_code_version_with_github_uri(uri, script_relative_path, login_or_token, password, jwt)

    @classmethod
    def create_code_version_with_gitlab_uri(
        cls, uri: str, script_relative_path: str, private_token: Optional[str] = None, oauth_token: Optional[str] = None
    ) -> Optional[CodeVersionArtifact]:
        """
         Create a code artifact based on the Gitlab information relative to the given URI and relative path.

        Note: The URI given can include the tree you are working on. otherwise, the default repository branch will be used.

        sample uri:
            "https://gitlab.com/my-organization/my-repository" (no branch given so using default branch)
            https://gitlab.com/my-organization/my-repository/tree/my-current-branch (branch given is my-current-branch)

        sample script relative path:
            "file_name.txt" (no folder in path)
            "folder/file_name.txt" (folder and file in path)

        To access private repositories, you need to authenticate with your credentials.
        see https://https://docs.gitlab.com/ee/topics/authentication/

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param private_token: A real login or a personal access token
        :param oauth_token: The OAuth2 access token
        :return: A CodeVersion or None if the Gitlab repository was not found or is not accessible
        """
        return Artifacts.create_code_version_with_gitlab_uri(uri, script_relative_path, private_token, oauth_token)

    @classmethod
    def create_code_version_with_bitbucket_uri(
        cls,
        uri: str,
        script_relative_path: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> Optional[CodeVersionArtifact]:
        """
        Create a code artifact based on the Bitbucket information relative to the given URI and relative path.

        Note: The URI given can include the branch you are working on. otherwise, the default repository branch will be used.

        sample :
            https://bitbucket.org/workspace/project/ (no branch given so using default branch)
            https://bitbucket.org/workspace/project/src/branch (branch given is my-current-branch)

        To access private repositories, you need to authenticate with your credentials.
        see Bitbucket Cloud: https://atlassian-python-api.readthedocs.io/index.html

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param username: Bitbucket email
        :param password: Bitbucket password
        :return: A CodeVersion or None if the Bitbucket repository was not found or is not accessible
        """
        return Artifacts.create_code_version_with_bitbucket_uri(uri, script_relative_path, username, password)

    @classmethod
    def create_bigquery_dataset_version(
        cls, uri: str, description: Optional[str] = None
    ) -> Optional[DatasetMetadataArtifact]:
        """
        Create a bigquery dataset version using a uri. Creates a BigQuery resource attached to the dataset version. The uri is found in the google
        cloud console in the BigQuery tab.

        Example:
        Tables level : "https://console.cloud.google.com/bigquery?project=Project_Name&d=Dataset_Name&p=Project_Name&t=Table_Name&page=table"
        Dataset level: "https://console.cloud.google.com/bigquery?project=Project_Name&d=Dataset_Name&p=Project_Name&p=Project_Name&page=dataset&ws=!1m4!1m3!3m2!1sProject_Name!2sDataset_Name"

        :param uri: The url of the BigQuery dataset version that will be created
        :param description: The description of the dataset version
        :return: A DatasetMetadataArtifact that can be used in runs as an input or output
        """
        return Artifacts.create_bigquery_dataset(uri, description)

    @classmethod
    def create_gcs_dataset_version(
        cls, uri: Union[str, List[str]], description: Optional[str] = None
    ) -> Optional[DatasetMetadataArtifact]:
        """
        Create a Google Cloud Storage dataset version using an uri or List of uris. Creates a Google Cloud Storage resource attached to the dataset version. The uri is found in the google
        cloud console in the Google Cloud Storage tab.

        Example:
        File level :  bucket_name/folder_name/file_name.csv (Attaches the file)
        Folder level: bucket_name/folder_name (Attaches everything in the the folder)
        List : ['bucket_name/folder_name/file_name.csv', 'bucket_name/folder_name']

        :param uri: The uri or List of the Google Cloud Storage resources that will be used to create a dataset version
        :param description: The description of the dataset version
        :return: A DatasetMetadataArtifact that can be used in runs as an input or output
        """
        return Artifacts.create_gcs_dataset(uri, description)



from __future__ import annotations
from dataclasses import dataclass, asdict
from datetime import datetime
from typing import Optional
import inspect
from vectice.models import metric


@dataclass
class __Output:
    key: str
    value: float
    timestamp: datetime
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, metrics):
        return cls(**{k: v for k, v in metrics.items() if k in inspect.signature(cls).parameters})

    def as_dict(self):
        return asdict(self)


@dataclass
class EntityMetricOutput(metric.Metric, __Output):
    pass



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from vectice.models import Job


@dataclass
class __Output:
    createdDate: datetime
    updatedDate: datetime
    version: int
    id: int
    workspaceId: int
    projectId: int
    authorId: int
    metadataSource: str
    deletedDate: Optional[datetime]


@dataclass
class JobOutput(Job, __Output):
    pass



from vectice.entity.job import Job
from vectice.entity.job_run import Run
from vectice.entity.job_artifact import Artifact
from vectice.entity.dataset import Dataset
from vectice.entity.dataset_version import DatasetVersion
from vectice.entity.model import Model
from vectice.entity.model_version import ModelVersion
from vectice.entity.code import Code
from vectice.entity.code_version import CodeVersion
from vectice.entity.git_version import GitVersion

__all__ = [
    "Job",
    "Run",
    "Artifact",
    "Dataset",
    "DatasetVersion",
    "Model",
    "ModelVersion",
    "Code",
    "CodeVersion",
    "GitVersion",
]



from typing import Optional
from urllib.parse import urlencode

from vectice.api.project import ProjectApi
from vectice.entity import Model

from .Page import Page
from .json_object import JsonObject
from .output.model_output import ModelOutput
from .output.paged_response import PagedResponse


class ModelApi(ProjectApi):
    def __init__(self, project_token: str, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._model_path = super().api_base_path + "/model"

    @property
    def api_base_path(self) -> str:
        return self._model_path

    def list_models(self, search: str = None, page_index=Page.index, page_size=Page.size) -> PagedResponse[ModelOutput]:
        queries = {"index": page_index, "size": page_size}
        if search:
            queries["search"] = search
        models = self._get(self.api_base_path + "?" + urlencode(queries))
        return PagedResponse(item_cls=ModelOutput, total=models["total"], page=models["page"], items=models["items"])

    def create_model(self, model: JsonObject) -> Model:
        if model.get("name") is None:
            raise ValueError('"name" must be provided in model.')
        if model.get("type") is None:
            model["type"] = "OTHER"
        return Model(self._post(self.api_base_path, model))

    def update_model(self, model_id: int, model: JsonObject) -> Model:
        return Model(self._put(self.api_base_path + "/" + str(model_id), model))



from typing import List, Optional
from urllib.parse import urlencode

from vectice.entity import Run

from .Page import Page
from .job import JobApi
from .json_object import JsonObject
from .output.job_run_output import JobRunOutput


class RunApi(JobApi):
    def __init__(self, project_token: str, job_id: int, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._job_id = job_id
        self._run_path = super().api_base_path + "/" + str(job_id) + "/run"

    @property
    def job_id(self) -> int:
        return self._job_id

    @property
    def api_base_path(self) -> str:
        return self._run_path

    def list_runs(self, page_index=Page.index, page_size=Page.size) -> List[JobRunOutput]:
        queries = {"index": page_index, "size": page_size}
        runs = self._get(self.api_base_path + "?" + urlencode(queries))["items"]
        return [JobRunOutput(**run) for run in runs]

    def create_run(self, run: JsonObject) -> Run:
        if run.get("status") is None:
            raise ValueError('"status" must be provided in run.')
        if run.get("startDate") is None:
            raise ValueError('"startDate" must be provided in run.')
        return Run(self._post(self.api_base_path, run))

    def update_run(self, run_id: int, run: JsonObject) -> Run:
        return Run(self._put(self.api_base_path + "/" + str(run_id), run))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, List, TypeVar, Tuple, Generic

from .with_parent import WithParent
from .with_tags import WithTags


class _Base(WithTags, WithParent):
    pass


V = TypeVar("V", bound="_Base")


class Artifact(ABC, Generic[V]):
    """
    Base class for all type of artifact
    """

    artifactType: str
    """"""
    jobArtifactType: Optional[str] = None
    """"""
    description: Optional[str] = None
    """"""

    T = TypeVar("T", bound="Artifact")

    @abstractmethod
    def _get_delegate(self) -> V:
        pass

    def __str__(self):
        return f"{self.__class__.__name__}({self.jobArtifactType},{str(self._get_delegate())})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.jobArtifactType},{str(self._get_delegate())})"

    def with_tag(self: T, key: str, value: str) -> T:
        """
        Add a tag to the artifact

        :param key: The key of the tag
        :param value: The value of the tag
        :return: itself
        """
        self._get_delegate().with_tag(key, value)
        return self

    def with_tags(self: T, tags: List[Tuple[str, str]]) -> T:
        """
        Add several tags to the artifact

        :param tags: A list of tuple `(key,value)`
        :return: itself
        """
        self._get_delegate().with_tags(tags)
        return self

    def with_parent_id(self: T, parent_id: int) -> T:
        """
        Links the artifact to the parent.

        It is the id of the model for a model_version,
        the id of the dataset for a dataset_version
        and the id of the repository for a code_version

        :param parent_id: The identifier of the parent
        :return: itself
        """
        self._get_delegate().with_parent_id(parent_id)
        return self

    def with_parent_name(self: T, parent_name: str) -> T:
        """
        Links the artifact to the parent.

        It is the name of the model for a model_version,
        the name of the dataset for a dataset_version
        and the name of the repository for a code_version

        :param parent_name: The name of the parent
        :return: itself
        """
        self._get_delegate().with_parent_name(parent_name)
        return self



from vectice.entity._base_entity import BaseEntity


class VersionedEntity(BaseEntity):
    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def created_date(self) -> str:
        return str(self["createdDate"])

    @property
    def updated_date(self) -> str:
        return str(self["updatedDate"])

    @property
    def deleted_date(self) -> str:
        return str(self["deletedDate"])



from .code_version_artifact import CodeVersionArtifact
from .artifact import Artifact
from .artifact_type import ArtifactType
from .artifact_version import ArtifactVersion
from .artifacts import Artifacts
from .code_version import CodeVersion
from .dataset_version import DatasetVersion
from .dataset_version_artifact import DatasetVersionArtifact
from .errors import VecticeError
from .git_version import GitVersion
from .job import Job
from .job_artifact_type import JobArtifactType
from .job_run import JobRun
from .job_run_status import JobRunStatus
from .job_type import JobType
from .metric import Metric
from .model_version import ModelVersion
from .model_version_artifact import ModelVersionArtifact
from .property import Property
from .runnable_job import RunnableJob
from .tag import Tag
from .user_declared_version import UserDeclaredVersion
from .user_version import UserVersion
from .data_resource import DataResource
from .api_token import ApiToken
from .create_dataset import CreateDataset
from .attachments import Attachments
from .with_attachments import WithAttachments
from .update_model_version import UpdateModelVersion
from .attachments import AttachmentType
from .dataset_metadata_artifact import DatasetMetadataArtifact

__all__ = [
    "Artifacts",
    "Artifact",
    "ArtifactType",
    "ArtifactVersion",
    "CodeVersion",
    "CodeVersionArtifact",
    "DatasetVersion",
    "DatasetVersionArtifact",
    "GitVersion",
    "Job",
    "JobArtifactType",
    "JobRun",
    "JobRunStatus",
    "JobType",
    "Metric",
    "Property",
    "ModelVersion",
    "ModelVersionArtifact",
    "Tag",
    "RunnableJob",
    "UserVersion",
    "UserDeclaredVersion",
    "VecticeError",
    "DataResource",
    "ApiToken",
    "CreateDataset",
    "Attachments",
    "WithAttachments",
    "UpdateModelVersion",
    "AttachmentType",
    "DatasetMetadataArtifact",
]



import logging

from typing import Optional, Any, List


from vectice.adapter.adapter import Adapter
from vectice.adapter.factory import LibraryFactory
from vectice.models import (
    Artifact,
    RunnableJob,
    Job,
    JobRun,
)


logger = logging.getLogger(__name__)


# Import experiment into Vectice with all associated run (could be long !)
def save_job(project_token: str, experiment_name: str, lib: str):
    try:
        vectice = Vectice(project_token=project_token, lib=lib)
        vectice.save_job_and_associated_runs(experiment_name)
    except Exception:
        logger.exception(f"saving job {experiment_name} failed")


# Import run of experiment into Vectice
def save_after_run(
    project_token: str,
    run: Any,
    lib: Optional[str],
    inputs: Optional[List[Artifact]] = None,
    outputs: Optional[List[Artifact]] = None,
) -> Optional[int]:
    """
        Save in Vectice platform information relative to this run.
        The run object can be of several type depending on which
        lib you are using.

    :param project_token: The token of the project the job belongs to
    :param run: The run we want to save
    :param lib: Name of the lib you are using (for now, None or MLFlow)
    :param inputs: A list of inputs (artifacts) you are using in this run
    :param outputs: A list of outputs (artifacts) you are using in this run
    :return: An identifier of the saved run or None if the run can not be saved
    """
    try:
        vectice = Vectice(project_token, lib)
        return vectice.save_run(run, inputs, outputs)
    except Exception:
        logger.exception("saving run failed")
        return None


def create_run(job_name: str, job_type: Optional[str] = None) -> RunnableJob:
    """
    Create a local object of a run. Note that the run is not created in
    Vectice server (and as a consequence is NOT visible until saved after the run).

    This object will save any information relative to a run and its associated job.

    The returned instance needs to be used with associated :func:`save_after_run`

    For job types, take a look at the list in :class:`~vectice.models.JobType`

    :param job_name: The name of the job involved in the run
    :param job_type: The type of the job involved in the run
    :return: A runnable job
    """
    if job_name is None:
        raise RuntimeError("Name of job must be provided.")
    job = Job(job_name)
    if job_type is not None:
        job.with_type(job_type)
    return RunnableJob(job, JobRun())


class Vectice(Adapter):
    """
    High level class to list jobs and runs but also save runs

    The ability to toggle the autocode feature is enabled with the autocode argument. If autocode=True,
    Vectice will capture code artifacts relative to the .git file where you are executing your code. More can be found
    in the documentation at doc.vectice.com
    """

    def __new__(cls, project_token: str, lib: str = None, *args, **kwargs):
        if lib is not None:
            return LibraryFactory.get_library(project_token, lib, *args, **kwargs)
        else:
            return super().__new__(cls)

    def __init__(
        self, project_token: str, lib: Optional[str] = None, autocode: bool = False, allow_self_signed_certificate=True
    ):
        super().__init__(
            project_token=project_token, autocode=autocode, allow_self_signed_certificate=allow_self_signed_certificate
        )

    @classmethod
    def prepare_run(cls, job_name: str, job_type: Optional[str] = None) -> RunnableJob:
        """
        Create a local object of a run. Note that the run is not created in
        Vectice server (and as a consequence is NOT visible).

        The returned instance need to be used with associated :func:`~Vectice.save_after_run`

        :param job_name: The name of the job the run is related to
        :param job_type: The type of the job.
        :return: A runnable job
        """
        return create_run(job_name, job_type)

    @classmethod
    def save_after_run(
        cls,
        project_token: str,
        run: Any,
        lib: Optional[str],
        inputs: Optional[List[Artifact]] = None,
        outputs: Optional[List[Artifact]] = None,
    ) -> Optional[int]:
        """
        Save all run information in Vectice server

        :param project_token: The token of the project the job is belong to
        :param run: The run we want to save
        :param lib: The name of the lib you are using (for now, None or MLFlow)
        :param inputs: A list of inputs (artifact) you are using in this run
        :param outputs: A list of outputs (artifact) you are using in this run
        :return: An identifier of the saved run or None if the run can not be saved
        """
        return save_after_run(project_token, run, lib, inputs, outputs)



from __future__ import annotations

from dataclasses import dataclass, asdict
from typing import Optional, Union, List, Tuple

from .attachments import Attachments
from .with_properties import WithProperties


@dataclass
class UpdateModelVersion(WithProperties):
    algorithmName: Optional[str] = None
    description: Optional[str] = None
    data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None
    """"""

    def as_dict(self):
        return asdict(self)

    @staticmethod
    def create_attachments(attachments: Optional[List[str]]):
        if attachments is not None and len(attachments) >= 1:
            return Attachments("modelversion").with_attachments(attachments)
        return None

    def _create_with_property(self, data_properties: Tuple[str, str]):
        key, value = data_properties
        return self.with_property(key, value)

    def _create_with_properties(self, data_properties: List[Tuple[str, str]]):
        return self.with_properties(properties=data_properties)

    def add_properties(
        self, data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None
    ) -> UpdateModelVersion:
        if data_properties and isinstance(data_properties, list):
            self._create_with_properties(data_properties=data_properties)
        elif data_properties and isinstance(data_properties, tuple):
            self._create_with_property(data_properties=data_properties)
        return self



from typing import List, Optional
from dataclasses import dataclass
from enum import Enum


class DataType(Enum):
    String = "Text"
    Integer = "Integer"
    Float = "RealNumber"
    Float64 = "RealNumber"
    Boolean = "Boolean"
    Spacial = "Spacial"
    Binary = "Binary"
    Timestamp = "Timestamp"
    DateTime = "DateTime"
    Complex = "Complex"
    Unknown = "Unknown"


@dataclass
class SchemaColumn:
    name: Optional[str] = None
    description: Optional[str] = None
    dataType: Optional[DataType] = None
    length: Optional[int] = 0
    precision: Optional[int] = None
    scale: Optional[int] = 0
    isUnique: Optional[bool] = False
    default: Optional[str] = None
    nullable: Optional[bool] = True
    isPrimaryKey: Optional[bool] = False
    isForeignKey: Optional[bool] = False


@dataclass
class DataResourceSchema:
    type: str
    name: str
    description: str
    fileFormat: str
    columns: Optional[List[SchemaColumn]] = None



import logging
from datetime import datetime
from typing import Optional, Any, List, Tuple

import mlflow
from mlflow.entities import Experiment as MlflowExperiment, Run as MlflowRun
from mlflow.tracking import MlflowClient
from mlflow.entities import Run

from vectice.adapter.adapter import Adapter, ActiveRun
from vectice.api._utils import read_env
from vectice.models import Artifact, JobRun, Job, Metric, Property, RunnableJob, Artifacts

VECTICE_RUN_TAG = "vectice.run.id"
_STATUSES_MAP = {
    "RUNNING": "STARTED",
    "SCHEDULED": "SCHEDULED",
    "FINISHED": "COMPLETED",
    "FAILED": "FAILED",
    "KILLED": "ABORTED",
}

_STATUSES_VECTICE_TO_MLFLOW = {
    "STARTED": "RUNNING",
    "SCHEDULED": "SCHEDULED",
    "COMPLETED": "FINISHED",
    "FAILED": "FAILED",
    "ABORTED": "KILLED",
}

UNWANTED_TAGS = ["mlflow.log-model.history"]


class MlflowAdapter(Adapter):
    def __init__(self, project_token: str, autocode: bool = False, tracking_uri: str = None, registry_uri: str = None):
        super().__init__(project_token=project_token, autocode=autocode)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._mlflow_tracking_uri, self._mlflow_registry_uri = read_env("MLFLOW_TRACKING_URI", "MLFLOW_REGISTRY_URI")
        if tracking_uri is not None:
            self._mlflow_tracking_uri = tracking_uri
        if registry_uri is not None:
            self._mlflow_registry_uri = registry_uri

        if mlflow.get_tracking_uri() is None:
            mlflow.set_tracking_uri(self._mlflow_tracking_uri)
        if mlflow.get_registry_uri() is None:
            mlflow.set_registry_uri(self._mlflow_registry_uri)
        self._mlflow_client = MlflowClient(self._mlflow_tracking_uri, self._mlflow_registry_uri)

    @property
    def tracking_uri(self) -> Optional[str]:
        """MLflow tracking URI"""
        return self._mlflow_tracking_uri

    @tracking_uri.setter
    def tracking_uri(self, mlflow_tracking_uri: str):
        self._mlflow_tracking_uri = mlflow_tracking_uri
        mlflow.set_tracking_uri(mlflow_tracking_uri)
        self._mlflow_client = MlflowClient(mlflow_tracking_uri, self._mlflow_registry_uri)

    @property
    def registry_uri(self) -> Optional[str]:
        """MLflow registry URI"""
        return self._mlflow_registry_uri

    @registry_uri.setter
    def registry_uri(self, mlflow_registry_uri: str):
        self._mlflow_registry_uri = mlflow_registry_uri
        mlflow.set_registry_uri(mlflow_registry_uri)
        self._mlflow_client = MlflowClient(self._mlflow_tracking_uri, mlflow_registry_uri)

    # Fluent usage
    def create_run(self, name: str, job_type: Optional[str] = None, run_name: Optional[str] = None) -> RunnableJob:
        mlflow.set_experiment(name)
        return super().create_run(name, job_type, run_name)

    def start_run(
        self, runnable_job: Optional[RunnableJob] = None, inputs: Optional[List[Artifact]] = None
    ) -> ActiveRun:
        ml_run = mlflow.start_run()
        job, run, output_artifacts = self._extract_from_mlflow(ml_run, runnable_job=runnable_job)
        runnable_job = self.get_current_runnable_job(runnable_job)
        runnable_job.job = job
        runnable_job.run = run
        return super().start_run(runnable_job, inputs)

    def end_run(self, *args, **kwargs):
        if mlflow.active_run() is None:
            self._logger.warning("No active run found.")
            return None
        active_run = args[0]
        ml_run_id = mlflow.active_run().info.run_id
        kwargs["status"] = _STATUSES_VECTICE_TO_MLFLOW[kwargs["status"]]
        mlflow.end_run(**kwargs)
        ml_run = self._mlflow_client.get_run(ml_run_id)
        job, run, output_artifacts = self._extract_from_mlflow(ml_run)
        active_run.job.update(job.__dict__)
        active_run.run.update(run.__dict__)
        self._mlflow_client.set_tag(ml_run.info.run_id, VECTICE_RUN_TAG, active_run.run["id"])
        super().end_run(active_run, output_artifacts)

    # Sync-up functions from MLflow to Vectice

    def save_all_experiments(self) -> None:
        """Import all MLflow experiments along with all their runs into Vectice jobs."""
        experiments = self._mlflow_client.list_experiments()
        for ml_exp in experiments:
            self.__save_experiment(ml_exp)

    def save_experiment(self, experiment_id: str) -> None:
        """Import one MLflow experiment along with all its runs into Vectice job."""
        ml_exp = self._mlflow_client.get_experiment(experiment_id)
        self.__save_experiment(ml_exp)

    def save_experiment_by_name(self, name: str) -> None:
        """Import one MLflow experiment retrieved by its name into Vectice job."""
        ml_exp = self._mlflow_client.get_experiment_by_name(name)
        if ml_exp is None:
            raise ValueError(f"Experiment with name={name} does not exist.")
        self.__save_experiment(ml_exp)

    def save_run(
        self,
        run: Any,
        inputs: Optional[List[Artifact]] = None,
        outputs: Optional[List[Artifact]] = None,
    ) -> Optional[int]:
        """Import one MLflow run into Vectice job run."""
        if isinstance(run, (mlflow.ActiveRun, Run)):
            return self.__save_run(run, inputs, outputs)
        else:
            raise RuntimeError("Incompatible object provided.")

    def __save_experiment(self, experiment: MlflowExperiment) -> None:
        runs = self._mlflow_client.search_runs([experiment.experiment_id], order_by=["attributes.start_time DESC"])
        for run in runs:
            if not run.data.tags.get(VECTICE_RUN_TAG):
                self.__save_run(run)
            else:
                logging.info(f"cancelling save of run {run.info.run_id} as it is already saved in vectice")

    def __save_run(
        self,
        ml_run: mlflow.ActiveRun,
        inputs: Optional[List[Artifact]] = None,
        outputs: Optional[List[Artifact]] = None,
    ) -> Optional[int]:
        job, run, mapped_output_artifacts = self._extract_from_mlflow(ml_run)
        runnable_job = RunnableJob(job, run)
        active_run = super().start_run(runnable_job, inputs)
        if active_run is not None:
            run_id: Optional[int] = int(active_run.run["id"])
            self._mlflow_client.set_tag(ml_run.info.run_id, VECTICE_RUN_TAG, run_id)
        else:
            run_id = None
        if outputs is not None:
            mapped_output_artifacts.extend(outputs)
        super().end_run(active_run, mapped_output_artifacts)
        return run_id

    def _extract_from_mlflow(
        self, mlflow_run: MlflowRun, experiment: MlflowExperiment = None, runnable_job: Optional[RunnableJob] = None
    ) -> Tuple[Job, JobRun, List[Artifact]]:
        ml_exp = experiment or self._mlflow_client.get_experiment(mlflow_run.info.experiment_id)
        job = self._extract_job(ml_exp, runnable_job)
        run = self._extract_run(mlflow_run, runnable_job)
        output_artifacts = self._extract_outputs(mlflow_run, ml_exp)
        return job, run, output_artifacts

    @staticmethod
    def _extract_job(experiment: MlflowExperiment, runnable_job: Optional[RunnableJob] = None) -> Job:
        job_type = None
        if runnable_job:
            job_type = runnable_job.job.type
        return Job(experiment.name, type=job_type if job_type else None)

    @staticmethod
    def _extract_run(mlflow_run: MlflowRun, runnable_job: Optional[RunnableJob] = None) -> JobRun:
        tags = []
        run_name = None
        for tag in mlflow_run.data.tags.items():
            if tag[0] not in UNWANTED_TAGS:
                tags.append(Property(tag[0], tag[1]))
        if runnable_job:
            run_name = runnable_job.run.name
        return JobRun(
            startDate=datetime.fromtimestamp(mlflow_run.info.start_time / 1000),
            endDate=datetime.fromtimestamp(mlflow_run.info.end_time / 1000) if mlflow_run.info.end_time else None,
            status=_STATUSES_MAP[mlflow_run.info.status],
            name=run_name if run_name else None,
        ).with_extended_properties(tags)

    def _extract_outputs(self, mlflow_run: MlflowRun, experiment: MlflowExperiment) -> List[Artifact]:
        # Extract MLflow params and metrics as Vectice Model Version
        metrics = self._extract_metrics(mlflow_run)
        properties = self._extract_model_properties(mlflow_run)
        result = (
            Artifacts.create_model_version()
            .with_parent_name(experiment.name)
            .with_user_version(mlflow_run.info.run_id)
            .with_algorithm(mlflow_run.data.tags.get("estimator_name"))
            .with_extended_metrics(metrics)
            .with_extended_properties(properties)
        )
        return [result]

    def _extract_metrics(self, mlflow_run: MlflowRun) -> List[Metric]:
        metrics: List[Metric] = []
        for key in mlflow_run.data.metrics.keys():
            for m in self._mlflow_client.get_metric_history(mlflow_run.info.run_id, key):
                metrics.append(Metric(m.key, m.value, datetime.fromtimestamp(m.timestamp / 1000)))
        return metrics

    def _extract_model_properties(self, mlflow_run: MlflowRun) -> List[Property]:
        properties = []
        for prop in mlflow_run.data.params.items():
            properties.append(Property(prop[0], prop[1]))
        try:
            properties.append(
                Property(
                    "mlflow_url",
                    f"{mlflow.get_tracking_uri()}/#/experiments/{mlflow_run.info.experiment_id}/runs/{mlflow_run.info.run_id}",
                )
            )
        except Exception as e:
            self._logger.warning(f"Mlflow url failed to generate due to {e}")
        return properties



from vectice.entity._base_entity import BaseEntity


class Dataset(BaseEntity):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def job_artifact_type(self) -> str:
        return str(self["jobArtifactType"])

    @property
    def pattern(self) -> str:
        return str(self["pattern"])

    @property
    def is_pattern_base(self) -> bool:
        return bool(self["isPatternBase"])

    @property
    def connection_id(self) -> int:
        return int(self["connectionId"])

    @property
    def created_by_user_id(self) -> int:
        return int(self["createdByUserId"])

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])

    @property
    def creation_date(self) -> str:
        return str(self["creationDate"])

    @property
    def update_date(self) -> str:
        return str(self["updateDate"])

    @property
    def delete_date(self) -> str:
        return str(self["deleteDate"])

    @property
    def is_deleted(self) -> bool:
        return bool(self["isDeleted"])



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeVar, List, Optional, Union

from .files_metadata import TreeItem, TreeItemType
from .data_resource_schema import DataResourceSchema


class WithFilesMetadataTrait(ABC):
    T = TypeVar("T", bound="WithFilesMetadataTrait")

    @abstractmethod
    def with_metadata(
        self: T,
        name: str,
        id: str = None,
        parentId: Optional[str] = None,
        path: Optional[str] = None,
        type: Optional[Union[TreeItemType, str]] = None,
        size: Optional[int] = None,
        isFolder: Optional[bool] = None,
        children: Optional[List[TreeItem]] = None,
        uri: Optional[str] = None,
        metadata: Optional[DataResourceSchema] = None,
        digest: Optional[str] = None,
        itemCreatedDate: Optional[str] = None,
        itemUpdatedDate: Optional[str] = None,
    ) -> T:
        """
        :param name: the name of the file/s or folder/s
        :param id: the id of the dataset
        :param parentId: the parent id of the file/s or folder/s
        :param path: the path of the file/s or folder/s
        :param type: the type of the file/s or folder/s
        :param size: the size of the file/s or folder/s
        :param children: the children (other files under the file/s or folder/s)
        :param uri: the uri of the file/s or folder/s
        :param metadata: the DataResourceSchema of the file/s or folder/s
        :param digest: the hash of the file or folder
        :param itemCreatedDate: creation date of the file or folder, utc time format is used e.g 'YYYY-MM-DD'
        :param itemUpdatedDate: update date of the file or folder, utc time format is used e.g 'YYYY-MM-DD'

        :return: itself
        """
        pass


@dataclass
class WithFilesMetadata(WithFilesMetadataTrait):
    filesMetadata: Optional[List[TreeItem]] = None

    T = TypeVar("T", bound="WithFilesMetadata")

    def with_metadata(
        self: T,
        name: Optional[str] = None,
        id: Optional[str] = None,
        parentId: Optional[str] = None,
        path: Optional[str] = None,
        type: Optional[Union[TreeItemType, str]] = None,
        size: Optional[int] = None,
        isFolder: Optional[bool] = None,
        children: Optional[List[TreeItem]] = None,
        uri: Optional[str] = None,
        metadata: Optional[DataResourceSchema] = None,
        digest: Optional[str] = None,
        itemCreatedDate: Optional[str] = None,
        itemUpdatedDate: Optional[str] = None,
    ) -> T:
        self.filesMetadata = [
            TreeItem(
                name=name,
                id=id,
                parentId=parentId,
                path=path,
                type=type,
                size=size,
                isFolder=isFolder,
                children=children,  # type: ignore
                uri=uri,
                metadata=metadata,
                digest=digest,
                itemCreatedDate=itemCreatedDate,
                itemUpdatedDate=itemUpdatedDate,
            )
        ]
        return self


class WithDelegatedFilesMetadata(WithFilesMetadataTrait, ABC):
    T = TypeVar("T", bound="WithDelegatedFilesMetadata")

    @abstractmethod
    def _get_delegate(self) -> WithFilesMetadataTrait:
        pass

    def with_metadata(
        self: T,
        name: str,
        id: Optional[str] = None,
        parentId: Optional[str] = None,
        path: Optional[str] = None,
        type: Optional[Union[TreeItemType, str]] = None,
        size: Optional[int] = None,
        isFolder: Optional[bool] = None,
        children: Optional[List[TreeItem]] = None,
        uri: Optional[str] = None,
        metadata: Optional[DataResourceSchema] = None,
        digest: Optional[str] = None,
        itemCreatedDate: Optional[str] = None,
        itemUpdatedDate: Optional[str] = None,
    ) -> T:
        """
        The with metadata method enables the creation of a dataset with metadata from popular cloud storage providers and local files. The
        only required kwarg is name. The other kwargs can be leveraged as needed to create and track dataset versions.

        :param name: the name of the file/s or folder/s
        :param id: the id of the dataset
        :param parentId: the parent id of the file/s or folder/s
        :param path: the path of the file/s or folder/s
        :param type: the type of the file/s or folder/s
        :param size: the size of the file/s or folder/s
        :param children: the children (other files under the file/s or folder/s)
        :param uri: the uri of the file/s or folder/s
        :param metadata: the DataResourceSchema of the file/s or folder/s
        :param digest: the hash of the file or folder
        :param itemCreatedDate: creation date of the file or folder, utc time format is used e.g 'YYYY-MM-DD'
        :param itemUpdatedDate: update date of the file or folder, utc time format is used e.g 'YYYY-MM-DD'

        :return: itself
        """
        self._get_delegate().with_metadata(
            name=name,
            id=id,
            parentId=parentId,
            path=path,
            type=type,
            size=size,
            isFolder=isFolder,
            children=children,
            uri=uri,
            metadata=metadata,
            digest=digest,
            itemCreatedDate=itemCreatedDate,
            itemUpdatedDate=itemUpdatedDate,
        )
        return self



from typing import Optional, List, Union

from .code_version_artifact import CodeVersionArtifact
from .dataset_version_artifact import DatasetVersionArtifact
from .model_version_artifact import ModelVersionArtifact
from .dataset_metadata_artifact import DatasetMetadataArtifact
from .api_token import ApiToken


class Artifacts:
    """
    Factory class for Artifacts
    """

    @classmethod
    def create_dataset_version(
        cls,
        description: Optional[str] = None,
    ) -> DatasetVersionArtifact:
        """
        Create an artifact for a dataset

        :param description: The description of the dataset version
        :return: A DatasetVersionArtifact
        """
        return DatasetVersionArtifact.create(description)

    @classmethod
    def create_model_version(
        cls,
        description: Optional[str] = None,
    ) -> ModelVersionArtifact:
        """
        Create an artifact for a model

        :param description: The description of the model version
        :return: A ModelVersionArtifact
        """
        return ModelVersionArtifact.create(description)

    @classmethod
    def create_code_version(cls, path: str = ".") -> Optional[CodeVersionArtifact]:
        """
        Create a code artifact based on the git information relative to the given local path.

        :param path: The path to look for the git repository
        :return: A CodeVersion or None if a git repository was not found locally
        """
        return CodeVersionArtifact.create(path)

    @classmethod
    def create_code_version_with_github_uri(
        cls, uri: str, script_relative_path: Optional[str] = None, login_or_token=None, password=None, jwt=None
    ) -> Optional[CodeVersionArtifact]:
        """
        Create a code artifact based on the github information relative to the given URI and relative path.

        Note: The URI given can include the branch you are working on. otherwise, the default repository branch will be used.

        sample :
            https://github.com/my-organization/my-repository (no branch given so using default branch)
            https://github.com/my-organization/my-repository/tree/my-current-branch (branch given is my-current-branch)

        To access private repositories, you need to authenticate with your credentials.
        see https://docs.github.com/en/github/authenticating-to-github/keeping-your-account-and-data-secure/about-authentication-to-github

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param login_or_token: A real login or personal access token
        :param password: The password
        :param jwt: The Oauth2 access token
        :return: A CodeVersion or None if the github repository was not found or is not accessible
        """
        return CodeVersionArtifact.create_from_github_uri(uri, script_relative_path, login_or_token, password, jwt)

    @classmethod
    def create_code_version_with_gitlab_uri(
        cls, uri: str, script_relative_path: str, private_token: Optional[str] = None, oauth_token: Optional[str] = None
    ) -> Optional[CodeVersionArtifact]:
        """
        Create an artifact that contains a version of a code

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param private_token: A real login or a personal access token
        :param oauth_token: The OAuth2 access token
        :return: A CodeVersion or None if the GitHub repository was not found or is not accessible
        """
        return CodeVersionArtifact.create_from_gitlab_uri(uri, script_relative_path, private_token, oauth_token)

    @classmethod
    def create_code_version_with_bitbucket_uri(
        cls,
        uri: str,
        script_relative_path: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> Optional[CodeVersionArtifact]:
        """
        Create a code artifact based on the Bitbucket information relative to the given URI and relative path.

        Note: The URI given can include the branch you are working on. otherwise, the default repository branch will be used.

        sample :
            https://bitbucket.org/workspace/project/ (no branch given so using default branch)
            https://bitbucket.org/workspace/project/src/branch (branch given is my-current-branch)

        To access private repositories, you need to authenticate with your credentials.
        see Bitbucket Cloud: https://atlassian-python-api.readthedocs.io/index.html

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param username: Bitbucket email
        :param password: Bitbucket password
        :return: A CodeVersion or None if the Bitbucket repository was not found or is not accessible
        """
        return CodeVersionArtifact.create_from_bitbucket_uri(uri, script_relative_path, username, password)

    @classmethod
    def parse_api_token(cls, token_file: str) -> ApiToken:
        """
        Parses the the API Token json file and sets the os environmental variable for the "VECTICE_API_TOKEN"
        for the user.

        :param token_file: The filepath to the json file containing the "VECTICE_API_TOKEN", found on the Vectice App
        :return: ApiToken or None if the json file can not be parsed or found
        """
        return ApiToken.parse_api_token_json(token_file)

    @classmethod
    def create_bigquery_dataset(cls, uri: str, description: Optional[str] = None) -> Optional[DatasetMetadataArtifact]:
        """
        Create a bigquery dataset version using an uri. Creates a BigQuery resource attached to the dataset version. The url is found in the google
        cloud console in the BigQuery tab.

        Example:
        Tables level : "https://console.cloud.google.com/bigquery?project=Project_Name&d=Dataset_Name&p=Project_Name&t=Table_Name&page=table"
        Dataset level: "https://console.cloud.google.com/bigquery?project=Project_Name&d=Dataset_Name&p=Project_Name&p=Project_Name&page=dataset&ws=!1m4!1m3!3m2!1sProject_Name!2sDataset_Name"

        :param uri: The url of the BigQuery dataset version that will be created
        :param description: The description of the dataset version
        :return: A DatasetMetadataArtifact that can be used in runs as an input or output
        """
        return DatasetMetadataArtifact.create_bigquery(uri, description)

    @classmethod
    def create_gcs_dataset(
        cls, uri: Union[str, List[str]], description: Optional[str] = None
    ) -> Optional[DatasetMetadataArtifact]:
        """
        Create a Google Cloud Storage dataset version using an uri. Creates a Google Cloud Storage resource attached to the dataset version. The uri is found in the google
        cloud console in the Google Cloud Storage tab.

        Example:
        File level :  bucket_name/folder_name/file_name.csv (Attaches the file)
        Folder level: bucket_name/folder_name (Attaches everything in the the folder)
        List : ['bucket_name/folder_name/file_name.csv', 'bucket_name/folder_name']

        :param uri: The uri list of the Google Cloud Storage resources that will be used to create a dataset version
        :param description: The description of the dataset version
        :return: A DatasetMetadataArtifact that can be used in runs as an input or output
        """
        return DatasetMetadataArtifact.create_gcs(uri, description)



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeVar


class WithAPITrait(ABC):
    T = TypeVar("T", bound="WithAPITrait")

    @abstractmethod
    def with_create(self: T) -> T:
        """

        :param key:
        :param value:
        :return: iself
        """
        pass


@dataclass
class WithAPI(WithAPITrait):

    T = TypeVar("T", bound="WithAPI")

    def with_create(self: T) -> T:
        from vectice.api import Client  # import here to avoid circular import
        import time

        payload = self.__dict__
        payload["name"] = payload.pop("parentName")
        dataset_id = Client(os.environ["PROJECT_TOKEN"]).create_dataset(payload)
        time.sleep(1)

        Client(os.environ["PROJECT_TOKEN"]).create_dataset_version(dataset_id["id"], self)  # type: ignore
        return self


class WithDelegatedAPI(WithAPITrait, ABC):
    T = TypeVar("T", bound="WithDelegatedAPI")

    @abstractmethod
    def _get_delegate(self) -> WithAPITrait:
        pass

    def with_create(self: T) -> T:
        self._get_delegate().with_create()
        return self



import json
import logging
import os
from base64 import b64decode
from datetime import datetime
from typing import Optional

from ._http import Connection
from ._utils import read_env


class Auth(Connection):  # nosec B107
    def __init__(
        self,
        api_endpoint: str = None,
        api_token: str = None,
        _token: Optional[str] = None,
        auto_connect=True,
        allow_self_certificate=True,
    ):
        self._logger = logging.getLogger("vectice.auth")
        endpoint, self._API_TOKEN = read_env("VECTICE_API_ENDPOINT", "VECTICE_API_TOKEN")
        if api_endpoint is not None:
            endpoint = api_endpoint
        if api_token is not None:
            self._API_TOKEN = api_token
        if not self._API_TOKEN:
            raise ValueError("VECTICE_API_TOKEN is not provided.")
        super().__init__(api_endpoint=endpoint, allow_self_certificate=allow_self_certificate)
        self._jwt = None
        self._jwt_expiration = None
        if _token:
            self._token = _token
        elif auto_connect:
            self._refresh_token()
        self.vectice_path: Optional[str] = None

    @property
    def _token(self) -> Optional[str]:
        if self._jwt_expiration is None:
            return None
        # Refresh token 1 min before expiration
        if datetime.now().timestamp() >= self._jwt_expiration - 60:
            self._refresh_token()
        return self._jwt

    @_token.setter
    def _token(self, jwt: str) -> None:
        self._jwt = jwt
        self._jwt_expiration = self._get_jwt_expiration(jwt)
        self._request_headers = {"Authorization": "Bearer " + jwt}

    def _refresh_token(self) -> None:
        self._logger.info("Vectice: Refreshing token... ")
        response = self._post("/metadata/authenticate", {"apiKey": self._API_TOKEN})
        self._token = response["token"]
        self._logger.info("Success!")

    def auth_project_token(self, project_token) -> None:
        self._logger.info("Vectice: Validating project token... ")
        response = self._get(f"/metadata/project/{project_token}/validate/")
        self._logger.info(
            f"The entered token is OK, and allows you to work on the '{response['name']}' Project, part of the '{response['workspace']['name']}' Workspace"
        )
        self.vectice_path = f"All your work is being captured in the '{response['name']}' Project, part of the '{response['workspace']['name']}' Workspace"

    @staticmethod
    def auth_api_token(api_token: str, api_endpoint: Optional[str]):
        logging.info("Vectice: Validating api token... ")
        if api_endpoint is None and os.getenv("VECTICE_API_ENDPOINT"):
            api_endpoint = os.getenv("VECTICE_API_ENDPOINT")
        elif api_endpoint is None and not os.getenv("VECTICE_API_ENDPOINT"):
            raise ValueError("No api endpoint could be found to connect to the Vectice App!")
        try:
            Connection(api_endpoint)._post("/metadata/authenticate", {"apiKey": api_token})
        except Exception as e:
            raise RuntimeError(f"Could not authenticate due to: {e}")
        logging.info("Success!")

    @staticmethod
    def _get_jwt_expiration(jwt: str) -> int:
        jwt_payload = jwt.split(".")[1]
        jwt_payload_with_padding = f"{jwt_payload}{'=' * (4 - len(jwt_payload) % 4)}"
        return int(json.loads(b64decode(jwt_payload_with_padding))["exp"])

    def token(self) -> Optional[str]:
        return self._token

    def connect(self) -> None:
        self._refresh_token()



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional


@dataclass
class Property:
    key: str
    """"""
    value: str
    """"""
    timestamp: datetime = field(default_factory=datetime.now)
    """"""
    name: Optional[str] = None
    """"""



from dataclasses import dataclass


@dataclass
class __Output:
    fileId: int
    fileName: str
    contentType: str
    entityId: int
    entityType: str


@dataclass
class AttachmentOutput(__Output):
    pass



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
import logging
import json


@dataclass
class ApiToken:
    name: Optional[str] = None
    description: Optional[str] = None
    key: str = ""

    @staticmethod
    def parse_api_token_json(token_file: str) -> ApiToken:
        logging.info("Parsing api token json file...")
        with open(token_file) as file:
            data = json.load(file)
        return ApiToken(**data)



import os
import json
import urllib.error
import urllib.request
import requests
from urllib.parse import unquote
from itertools import chain
from pathlib import Path, PurePath
from typing import Generator, Tuple, Union


FILE_ERROR = "Can't identify the notebook {}. When attempting to create auto code artifact."
CONN_ERROR = "Unable to access server;\n" + "ipynbname requires either no security or token based security."


class HttpError(Exception):
    def __init__(self, code: int, reason: str, path: str):
        super().__init__()
        self.code: int = code
        self.reason: str = reason
        self.path = path

    def __str__(self):
        return f"""HTTP Error Code {self.code} : {self.reason}
        {self.path}
        """


def _list_running_servers(runtime_dir=None) -> Generator[dict, None, None]:
    """Iterate over the server info files of running notebook servers."""
    from jupyter_core.paths import jupyter_runtime_dir

    if runtime_dir is None:
        runtime_dir = jupyter_runtime_dir()
    runtime_dir = Path(runtime_dir)

    if runtime_dir.is_dir():
        for file_name in chain(
            runtime_dir.glob("nbserver-*.json"),  # jupyter notebook (or lab 2)
            runtime_dir.glob("jpserver-*.json"),  # jupyterlab 3
        ):
            yield json.loads(file_name.read_bytes())


def _get_kernel_id() -> str:
    """Returns the kernel ID of the ipykernel."""
    import ipykernel

    connection_file = Path(ipykernel.get_connection_file()).stem
    kernel_id = connection_file.split("-", 1)[1]
    return kernel_id


def _get_sessions(server):
    """Given a server, returns sessions, or HTTPError if access is denied.
    NOTE: Works only when either there is no security or there is token
    based security. An HTTPError is raised if unable to connect to a
    server.
    """
    try:
        query_str = ""
        token = server["token"]
        if token:
            query_str = f"?token={token}"
        url = f"{server['url']}api/sessions{query_str}"  # nosec
        with urllib.request.urlopen(url) as request:  # nosec
            return json.load(request)  # nosec
    except Exception:
        raise HttpError(code=request.status, reason=CONN_ERROR, path=url)


def _find_notebook_path() -> Union[Tuple[dict, PurePath], Tuple[None, None]]:
    from traitlets.config import MultipleInstanceError

    try:
        kernel_id = _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        return None, None  # Could not determine
    for server in _list_running_servers():
        try:
            sessions = _get_sessions(server)
            for sess in sessions:
                if sess["kernel"]["id"] == kernel_id:
                    return server, PurePath(sess["notebook"]["path"])  # nosec
        # There may be stale entries in the runtime directory
        except Exception:  # nosec
            pass  # nosec
    return None, None  # nosec


def notebook_name() -> str:
    """Returns the short name of the notebook w/o the .ipynb extension,
    or raises a FileNotFoundError exception if it cannot be determined.
    """
    _, path = _find_notebook_path()
    if path:
        return path.stem
    raise FileNotFoundError(FILE_ERROR.format("name"))


def notebook_path() -> Path:
    """Returns the absolute path of the notebook,
    or raises a FileNotFoundError exception if it cannot be determined.
    """
    srv, path = _find_notebook_path()
    if srv and path:
        root_dir = Path(srv.get("root_dir") or srv["notebook_dir"])
        return root_dir / path
    raise FileNotFoundError(FILE_ERROR.format("path"))


def auth_drive():
    """Authenticate Google Drive"""
    from pydrive.auth import GoogleAuth
    from pydrive.drive import GoogleDrive
    from oauth2client.client import GoogleCredentials as GC
    from google.oauth2.credentials import Credentials
    from googleapiclient.discovery import build
    from google.colab import auth

    SCOPES = ["https://www.googleapis.com/auth/drive.metadata.readonly"]
    drive = None
    if os.getenv("GOOGLE_DRIVE_TOKEN"):
        creds = Credentials.from_authorized_user_file(os.getenv("GOOGLE_DRIVE_TOKEN"), SCOPES)
        drive = build("drive", "v3", credentials=creds, cache_discovery=False)
    elif os.getenv("GOOGLE_DRIVE_TOKEN") is None:
        auth.authenticate_user()
        gauth = GoogleAuth()
        gauth.credentials = GC.get_application_default()
        drive = GoogleDrive(gauth)
    return drive


def get_current_collab_notebook():
    """Get the filename, file_id of this notebook"""
    d = requests.get("http://172.28.0.2:9000/api/sessions").json()[0]
    filename = unquote(d["name"])
    file_id = d["path"].split("=")[1]
    return filename, file_id


def get_collab_path(file_id, drive=None):
    """
    Use file_id to get the file path in Google Drive

    NB needs to be mounted. User setup is conditional:

    from google.colab import drive
    drive.mount('/content/drive')
    """
    if drive is None:
        drive = auth_drive()
    f = drive.CreateFile({"id": file_id})
    name = f["title"]
    if f["parents"]:
        parent_id = f["parents"][0]["id"]
        return get_collab_path(parent_id, drive) / name
    else:
        return Path(name)


def _find_git_path(drive=None):
    """
    Find the git path relative to the google drive directory.
    """
    if drive is None:
        drive = auth_drive()
    file_list = drive.ListFile({"q": " title='.git' "}).GetList()
    return get_collab_path(file_list[0]["id"])


def get_absolute_path() -> str:
    """
    Get the path for the directory walk to check if there's a .git and initialize the Repo object.
    Relates to 'drive.mount('/content/drive')' as '/content/' is part of the path and 'drive/' is
    required or the path will not be found.
    """
    filename, file_id = get_current_collab_notebook()
    path = str(get_collab_path(file_id=file_id, drive=None))
    try:
        path_formatting = path.split("/")
        path_formatting.pop()
        completed_path = "/".join(path_formatting)
        return f"drive/{completed_path}/{filename}"
    except ValueError:
        return str(path)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from vectice.models import model


@dataclass
class __Output:
    createdDate: datetime
    updatedDate: datetime
    version: int
    id: int
    authorId: int
    workspaceId: int
    projectId: int
    repository: str
    deletedDate: Optional[datetime]


@dataclass
class ModelOutput(model.Model, __Output):
    pass



from datetime import datetime, timezone
from typing import List, Optional

from vectice.models import Job, Artifact, JobRun, VecticeError

from ._http import HttpError
from ._utils import calculate_duration
from .json_object import JsonObject
from .project import ProjectApi
from ..entity import DatasetVersion, ModelVersion, CodeVersion
from ..entity.job_artifact import ArtifactType


class RuleApi(ProjectApi):
    def __init__(self, project_token: str, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._rule_path = super().api_base_path + "/rules"

    @property
    def api_base_path(self) -> str:
        return self._rule_path

    def start_run(self, job: Job, run: Optional[JobRun] = None, inputs: Optional[List[Artifact]] = None) -> JsonObject:
        if run is None:
            run = JobRun()
        if inputs is None:
            inputs = []
        else:
            inputs = [item for item in inputs if item is not None]
        if job.name is None:
            raise ValueError('"name" must be provided in job.')
        # self._check_artifacts(inputs)
        return self._post(
            self.api_base_path + "/start-run",
            {"job": job, "jobRun": run, "inputArtifacts": inputs},
        )

    def stop_run(self, run: dict, outputs: Optional[List[Artifact]] = None):
        if run.get("id") is None:
            raise ValueError('"id" must be provided in run.')
        if run.get("endDate") is None:
            now = datetime.now(timezone.utc)
            run["endDate"] = now.isoformat()
            run["duration"] = calculate_duration(now, run["startDate"])
        if outputs is None:
            outputs = []
        else:
            outputs = [item for item in outputs if item is not None]
        # self._check_artifacts(outputs)
        try:
            return self._post(self.api_base_path + "/stop-run", {"jobRun": run, "outputArtifacts": outputs})
        except HttpError as e:
            raise VecticeError("stop_run", e)

    @staticmethod
    def _check_artifacts(artifacts: List[dict]):
        for i in range(len(artifacts)):
            artifact = artifacts[i]
            # Fill DTO from an existing artifact
            if isinstance(artifact, DatasetVersion):
                artifacts[i] = {
                    "artifactType": "DATASET",
                    "dataset": {
                        "parentId": artifact.dataset_id,
                        "autoVersion": False,
                        "version": {"versionNumber": artifact.version_number},
                    },
                }
                continue
            if isinstance(artifact, ModelVersion):
                artifacts[i] = {
                    "artifactType": "MODEL",
                    "model": {
                        "parentId": artifact.model_id,
                        "version": {"versionNumber": artifact.version_number},
                    },
                }
                continue
            if isinstance(artifact, CodeVersion):
                artifacts[i] = {
                    "artifactType": "CODE",
                    "code": {
                        "parentId": artifact.code_id,
                        "version": {"versionNumber": artifact.version_number},
                    },
                }
                continue
            if "artifactType" not in artifact:
                raise ValueError('"artifactType" must be provided in artifact.')
            if artifact["artifactType"] not in ArtifactType.__members__:
                raise ValueError(f"Unsupported artifact type: {artifact['artifactType']}")



from dataclasses import dataclass


@dataclass
class Page:
    """
    simple structure to allow paging when requesting list of elements
    """

    index: int = 1
    """
    the index of the page
    """
    size: int = 20
    """
    the size of the page.
    """



from vectice.entity._user_declared_version import UserDeclaredVersion


class CodeVersion(UserDeclaredVersion):
    @property
    def code_id(self) -> int:
        return int(self["codeId"])

    @property
    def git_version_id(self) -> int:
        return int(self["gitVersionId"])



from dataclasses import dataclass
from typing import Optional


@dataclass
class UserDeclaredVersion:
    name: Optional[str] = None
    """"""
    uri: Optional[str] = None
    """"""
    description: Optional[str] = None
    """"""



from enum import Enum

from vectice.entity._user_declared_version import UserDeclaredVersion


class ModelVersion(UserDeclaredVersion):
    @property
    def status(self) -> str:
        return str(self["status"])

    @property
    def algorithm_name(self) -> str:
        return str(self["algorithmName"])

    @property
    def model_id(self) -> int:
        return int(self["modelId"])

    @property
    def origin_id(self) -> int:
        return int(self["originId"])


class ModelVersionStatus(Enum):
    EXPERIMENTATION = "EXPERIMENTATION"
    PRODUCTION = "PRODUCTION"
    STAGING = "STAGING"



from enum import EnumMeta


class JobRunStatus(EnumMeta):
    """
    Indicates the status of a run
    """

    CREATED = "CREATED"
    """
"""
    STARTED = "STARTED"
    """
"""
    SCHEDULED = "SCHEDULED"
    """
"""
    COMPLETED = "COMPLETED"
    """
"""
    FAILED = "FAILED"
    """
"""
    ABORTED = "ABORTED"
    """
"""



from . import models
from .api.client import Client
from .vectice import Vectice, create_run, save_after_run, save_job

__all__ = ["Vectice", "create_run", "save_after_run", "save_job", "Client", "models"]



from vectice.entity._versioned_entity import VersionedEntity


class GitVersion(VersionedEntity):
    @property
    def repository_name(self) -> str:
        return str(self["repositoryName"])

    @property
    def branch_name(self) -> str:
        return str(self["branchName"])

    @property
    def commit_hash(self) -> str:
        return str(self["commitHash"])

    @property
    def commit_comment(self) -> str:
        return str(self["commitComment"])

    @property
    def commit_author_name(self) -> str:
        return str(self["commitAuthorName"])

    @property
    def commit_author_email(self) -> str:
        return str(self["commitAuthorEmail"])

    @property
    def is_dirty(self) -> bool:
        return bool(self["isDirty"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])



from enum import Enum

from vectice.entity._user_declared_version import UserDeclaredVersion


class DatasetVersion(UserDeclaredVersion):
    @property
    def version_type(self) -> str:
        return str(self["versionType"])

    @property
    def dataset_id(self) -> int:
        return int(self["dataSetId"])

    @property
    def version_folder_id(self) -> str:
        return str(self["versionFolderId"])

    @property
    def origin_id(self) -> int:
        return int(self["originId"])


class DatasetVersionType(Enum):
    UNKNOWN = "UNKNOWN"
    USER_DECLARED = "USER_DECLARED"
    AUTO_VERSION = "AUTO_VERSION"



from enum import Enum

from vectice.entity._versioned_entity import VersionedEntity


class Job(VersionedEntity):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])

    @property
    def metadata_source(self) -> str:
        return str(self["metadataSource"])


class JobType(Enum):
    EXTRACTION = "EXTRACTION"
    PREPARATION = "PREPARATION"
    TRAINING = "TRAINING"
    INFERENCE = "INFERENCE"
    DEPLOYMENT = "DEPLOYMENT"
    OTHER = "OTHER"



from __future__ import annotations

from typing import Optional

from .artifact import Artifact
from .artifact_type import ArtifactType
from .dataset_version import DatasetVersion
from .with_data_resources import WithDelegatedDataResources
from .with_properties import WithDelegatedProperties
from .with_version import WithDelegatedVersion
from .with_files_metadata import WithDelegatedFilesMetadata


class DatasetVersionArtifact(
    Artifact, WithDelegatedProperties, WithDelegatedVersion, WithDelegatedDataResources, WithDelegatedFilesMetadata
):
    """ """

    def __init__(self, dataset: DatasetVersion, description: Optional[str] = None):
        self.artifactType = ArtifactType.DATASET
        self.description = description
        self.dataset: DatasetVersion = dataset

    @classmethod
    def create(
        cls,
        description: Optional[str] = None,
    ) -> DatasetVersionArtifact:
        """ """
        return cls(DatasetVersion(), description)

    def _get_delegate(self) -> DatasetVersion:
        return self.dataset

    def with_auto_version(self, auto_version=True) -> DatasetVersionArtifact:
        """
        Indicate if we should or not use the auto version mechanism

        :return: itself
        """
        self.dataset.with_auto_version(auto_version)
        return self



from __future__ import annotations
from dataclasses import dataclass, asdict
from datetime import datetime
from typing import Optional
import inspect
from vectice.models import property


@dataclass
class __Output:
    id: int
    dataSetVersionId: int
    key: str
    value: str
    timestamp: datetime
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, properties):
        return cls(**{k: v for k, v in properties.items() if k in inspect.signature(cls).parameters})

    def as_dict(self):
        return asdict(self)


@dataclass
class EntityPropertyOutput(property.Property, __Output):
    pass



from enum import Enum

from vectice.entity._versioned_entity import VersionedEntity


class Model(VersionedEntity):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def repository(self) -> str:
        return str(self["repository"])

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])


class ModelType(Enum):
    ANOMALY_DETECTION = "ANOMALY_DETECTION"
    CLASSIFICATION = "CLASSIFICATION"
    CLUSTERING = "CLUSTERING"
    OTHER = "OTHER"
    RECOMMENDATION_MODELS = "RECOMMENDATION_MODELS"
    REGRESSION = "REGRESSION"
    TIME_SERIES = "TIME_SERIES"



from dataclasses import dataclass
from typing import Optional


@dataclass
class DataResourcePath:
    id: Optional[str]
    parentId: Optional[str]
    path: str
    uri: Optional[str]
    uniqueId: Optional[str]
    type: Optional[str]
    isFolder: Optional[bool]



from vectice.entity._versioned_entity import VersionedEntity


class Code(VersionedEntity):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def uri(self) -> str:
        return str(self["type"])

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])



from typing import Optional, Union, List
from urllib.parse import urlencode

from vectice.entity import ModelVersion

from .Page import Page
from .json_object import JsonObject
from .model import ModelApi
from .output.metric_output import EntityMetricOutput
from .output.model_version_output import ModelVersionOutput
from .output.paged_response import PagedResponse
from .output.property_output import EntityPropertyOutput


class ModelVersionApi(ModelApi):
    def __init__(self, project_token: str, model_id: int, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._model_id = model_id
        self._model_version_path = super().api_base_path + "/" + str(model_id) + "/version"

    @property
    def model_id(self) -> int:
        return self._model_id

    @property
    def api_base_path(self) -> str:
        return self._model_version_path

    def list_model_versions(self, page_index=Page.index, page_size=Page.size) -> PagedResponse[ModelVersionOutput]:
        queries = {"index": page_index, "size": page_size}
        model_versions = self._get(self.api_base_path + "?" + urlencode(queries))
        return PagedResponse(
            item_cls=ModelVersionOutput,
            total=model_versions["total"],
            page=model_versions["page"],
            items=model_versions["items"],
        )

    def create_model_version(self, model_version: JsonObject) -> ModelVersion:
        if model_version.get("status") is None:
            raise ValueError('"status" must be provided in model_version.')
        return ModelVersion(self._post(self.api_base_path, model_version))

    def update_model_version(self, model_id: int, model_version: JsonObject) -> ModelVersion:
        return ModelVersion(self._put(self.api_base_path + "/" + str(model_id), model_version))

    def create_model_version_properties(self, model_version_id: int, properties: Union[List[dict], dict]):
        if isinstance(properties, dict):
            ModelVersion(self._post(self.api_base_path + f"/{model_version_id}/entityProperty/", properties))
        elif len(properties) >= 1:
            for prop in properties:
                ModelVersion(self._post(self.api_base_path + f"/{model_version_id}/entityProperty/", prop))

    def update_model_version_properties(self, model_version_id: int, property_id: int, properties: JsonObject):
        return ModelVersion(
            self._put(self.api_base_path + f"/{model_version_id}/entityProperty/{property_id}", properties)
        )

    def list_model_version_metrics(
        self, model_version_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[EntityMetricOutput]:
        queries = {"index": page_index, "size": page_size}
        model_version_metrics = self._get(
            self.api_base_path + f"/{model_version_id}/entityMetric?" + urlencode(queries)
        )
        metrics = [EntityMetricOutput.from_dict(metric).as_dict() for metric in model_version_metrics["items"]]
        return PagedResponse(
            item_cls=EntityMetricOutput,
            total=model_version_metrics["total"],
            page=model_version_metrics["page"],
            items=metrics,
        )

    def list_model_version_properties(
        self, model_version_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[EntityPropertyOutput]:
        queries = {"index": page_index, "size": page_size}
        model_version_properties = self._get(
            self.api_base_path + f"/{model_version_id}/entityProperty?" + urlencode(queries)
        )
        properties = [
            EntityPropertyOutput.from_dict(property).as_dict() for property in model_version_properties["items"]
        ]
        return PagedResponse(
            item_cls=EntityPropertyOutput,
            total=model_version_properties["total"],
            page=model_version_properties["page"],
            items=properties,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from .artifact import _Base
from .with_data_resources import WithDataResources
from .with_properties import WithProperties
from .with_version import WithVersion
from .with_files_metadata import WithFilesMetadata


@dataclass
class DatasetVersion(_Base, WithVersion, WithProperties, WithDataResources, WithFilesMetadata):
    """ """

    autoVersion: bool = True

    def with_auto_version(self, auto_version=True) -> DatasetVersion:
        """ """
        self.autoVersion = auto_version
        self.userDeclaredVersion = None
        self.version = None
        return self

    def with_user_version(
        self,
        name: Optional[str] = None,
        uri: Optional[str] = None,
        description: Optional[str] = None,
    ) -> DatasetVersion:
        self.autoVersion = False
        return super().with_user_version(name, uri, description)

    def with_existing_version_number(self, version_number: int) -> DatasetVersion:
        self.autoVersion = False
        return super().with_existing_version_number(version_number)

    def with_existing_version_name(self, version_name: str) -> DatasetVersion:
        self.autoVersion = False
        return super().with_existing_version_name(version_name)

    def with_existing_version_id(self, version_id: int) -> DatasetVersion:
        self.autoVersion = False
        return super().with_existing_version_id(version_id)



from __future__ import annotations

from enum import Enum
from dataclasses import dataclass
from typing import Optional, List, Tuple, Any, BinaryIO
from .with_attachments import WithAttachments


@dataclass
class Attachments(WithAttachments):
    artifact_type: Optional[str] = None
    """"""
    files: Optional[List[Tuple[str, Tuple[Any, BinaryIO]]]] = None
    """"""


class AttachmentType(Enum):
    """
    Enumeration of supported artifact types.
    """

    MODELVERSION = "modelversion"
    """
    """

    CODEVERSION = "codeversion"
    """
    """

    DATASETVERSION = "datasetversion"
    """
    """



from __future__ import annotations

import imp
import logging
import os
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple, Dict

import requests
import json

from vectice.utils.ipykernal_hook import notebook_path, get_absolute_path, notebook_name

LOGGER = logging.getLogger("GitVersion")


def _is_git_repo(path: str = ".", search_parent_directories: bool = True) -> bool:
    from git import Repo, InvalidGitRepositoryError

    try:
        Repo(path, search_parent_directories=search_parent_directories)
        return True
    except InvalidGitRepositoryError:
        return False


def _main_is_frozen() -> bool:
    return (
        hasattr(sys, "frozen") or hasattr(sys, "importers") or imp.is_frozen("__main__")  # new py2exe  # old py2exe
    )  # tools/freeze


def _get_executable_path() -> str:
    if _main_is_frozen():
        # print 'Running from path', os.path.dirname(sys.executable)
        return sys.executable
    return sys.argv[0]


def _relative_path(parent_path, child_path) -> str:
    parent_path = Path(os.path.abspath(parent_path))
    child_path = Path(os.path.abspath(child_path))
    try:
        return str(child_path.relative_to(parent_path)).replace("\\", "/")
    except ValueError:
        return str(child_path)


def _extract_git_version(path: str = ".", search_parent_directories: bool = True) -> Optional[GitVersion]:
    try:
        from git import Repo, InvalidGitRepositoryError
    except ModuleNotFoundError:
        return None

    repo = None
    if _check_if_collab():
        try:
            path = get_absolute_path()
            repo = Repo(path=path, search_parent_directories=search_parent_directories)
        except Exception:
            return None

    try:
        if repo is None:
            repo = Repo(path, search_parent_directories=search_parent_directories)
        repository_name = repo.remotes.origin.url.split(".git")[0].split("/")[-1]
        try:
            branch_name = repo.active_branch.name
        except TypeError:
            return None
        commit_hash = repo.head.object.hexsha
        commit_comment = repo.head.object.message
        commit_author_name = repo.head.object.author.name
        commit_author_email = repo.head.object.author.email
        if "ipykernel_launcher.py" in _get_executable_path() and not _check_if_collab():
            entrypoint = _relative_path(os.path.dirname(repo.git_dir), notebook_path())
        # TODO clean up
        elif _check_if_collab():
            entrypoint = get_absolute_path()
            entrypoint = entrypoint.split(str(repo.common_dir).split("/")[-2])[-1]
            entrypoint = entrypoint[1:]
        else:
            entrypoint = _relative_path(os.path.dirname(repo.git_dir), _get_executable_path())
        try:
            repo.active_branch.commit.tree.join(entrypoint)
        except KeyError:
            return None
        is_dirty = repo.is_dirty()
        uri = repo.remotes.origin.url
        return GitVersion(
            repository_name,
            branch_name,
            commit_hash,
            commit_comment,
            commit_author_name,
            commit_author_email,
            is_dirty,
            uri,
            entrypoint,
        )
    except InvalidGitRepositoryError:
        return None


def _extract_github_information_from_uri(uri: str) -> Optional[Tuple[str, str, str]]:
    match = re.search("(?:https://github.com/|git@github.com:)([a-zA-Z0-9-]+)/([a-zA-Z0-9-]+)(?:/tree/(.+))?", uri)
    if match:
        return match.group(1), match.group(2), match.group(3)
    else:
        return None


def _extract_bitbucket_cloud_information_from_uri(uri: str) -> Optional[Tuple[str, str, str]]:
    match = re.search("(?:https://bitbucket.org/|git@bitbucket.org:)([a-zA-Z0-9-]+)/([a-zA-Z0-9-]+)(?:/src/(.+))?", uri)
    if match:
        return match.group(1), match.group(2), match.group(3)
    else:
        return None


def _extract_gitlab_information_from_uri(uri: str) -> Optional[Tuple[str, str, str]]:
    match = re.search("(?:https://gitlab.com/|git@gitlab.com:)([a-zA-Z0-9-]+)/([a-zA-Z0-9-]+)(?:/tree/(.+))?", uri)
    if match:
        return match.group(1), match.group(2), match.group(3)
    else:
        return None


def _create_git_version_from_uri(
    uri: str, script_relative_path: Optional[str] = None, login_or_token=None, password=None, jwt=None
):
    from github import Github

    g = Github(login_or_token=login_or_token, password=password, jwt=jwt)
    result = _extract_github_information_from_uri(uri)
    if result is None:
        return None
    organisation, repository_name, branch = result
    github_repository = g.get_repo(f"{organisation}/{repository_name}")
    if branch is None:
        branch_name = github_repository.default_branch
    else:
        branch_name = branch
    github_branch = github_repository.get_branch(branch_name)
    commit = github_branch.commit
    if github_branch is None:
        raise RuntimeError(f"invalid branch name {branch_name}")
    if script_relative_path and not _is_file_in_git(github_repository, script_relative_path):
        return None
    return GitVersion(
        repository_name,
        branch_name,
        commit.sha,
        commit.commit.message,
        commit.commit.author.name,
        commit.commit.author.email,
        False,
        f"https://www.github.com/{organisation}/{repository_name}",
        script_relative_path,
    )


def _create_bitbucket_version_from_uri(
    uri: str, script_relative_path: Optional[str] = None, username: Optional[str] = None, password: Optional[str] = None
):
    from atlassian.bitbucket.cloud import Cloud  # type: ignore

    result = _extract_bitbucket_cloud_information_from_uri(uri)
    if result is None:
        return None
    workspace_name, repository_name, branch = result
    cloud = Cloud(url="https://api.bitbucket.org/", username=username, password=password, cloud=True)
    try:
        workspace = cloud.workspaces.get(workspace_name)
        bit_repository = workspace.repositories.get(repository_name).data
    except Exception as e:
        raise RuntimeError(f"{e}")
    if branch is None:
        branch = bit_repository["mainbranch"]["name"]
    try:
        branches_link = bit_repository["links"]["branches"]["href"]
        response = (
            requests.get(branches_link, auth=(username, password))
            if username and password
            else requests.get(branches_link)
        )
        branches = json.loads(response.text)
        commit = [br for br in branches.get("values") if br.get("name") == branch][0]
    except Exception:
        raise RuntimeError(f"invalid branch name '{branch}'")
    check_file = (
        requests.get(
            f"https://bitbucket.org/{workspace_name}/{repository_name}/full-commit/{commit['target']['hash']}/{script_relative_path}",
            auth=(username, password),
        )
        if username and password
        else requests.get(
            f"https://bitbucket.org/{workspace_name}/{repository_name}/full-commit/{commit['target']['hash']}/{script_relative_path}"
        )
    )
    if check_file.status_code > 200:
        return None
    return GitVersion(
        repository_name,
        branch,
        commit["target"]["hash"],
        commit["target"]["message"],
        commit["target"]["author"]["user"]["display_name"],
        commit["target"]["author"]["raw"],
        False,
        f"https://bitbucket.org/{workspace_name}/{repository_name}/",
        f"{script_relative_path}?at={branch}",
    )


def _create_gitlab_version_from_uri(
    uri: str, script_relative_path: str, private_token: Optional[str] = None, oauth_token: Optional[str] = None
):
    from gitlab import Gitlab

    job_token = os.environ.get("CI_JOB_TOKEN") if os.environ.get("CI_JOB_TOKEN") else None
    glab = Gitlab("https://gitlab.com/", private_token=private_token, oauth_token=oauth_token, job_token=job_token)
    result = _extract_gitlab_information_from_uri(uri)
    if result is None:
        return None
    organisation, repository_name, branch = result
    gitlab_repository = glab.projects.get(f"{organisation}/{repository_name}")
    if branch is None:
        branch_name = gitlab_repository.attributes.get("default_branch")
    else:
        branch_name = branch
    gitlab_branch = gitlab_repository.files.blame(file_path=script_relative_path, ref=branch_name)
    gitlab_file = gitlab_repository.files.get(file_path=script_relative_path, ref=branch_name)
    commit: Dict[str, str] = [
        curr_commit.get("commit")  # type: ignore
        for curr_commit in gitlab_branch
        if curr_commit.get("commit").get("id") == gitlab_file.last_commit_id  # type: ignore
    ][0]
    if gitlab_branch is None:
        raise RuntimeError(f"invalid branch name {branch_name}")
    return GitVersion(
        repository_name,
        branch_name,
        commit["id"],
        commit["message"],
        commit["author_name"],
        commit["author_email"],
        False,
        f"https://www.gitlab.com/{organisation}/{repository_name}",
        script_relative_path,
    )


def _is_file_in_git(github_repository=None, script_relative_path: Optional[str] = None) -> bool:
    if not script_relative_path:
        return False
    elif script_relative_path and github_repository:
        try:
            script_relative_path = script_relative_path.replace("%20", " ")
            github_repository.get_contents(script_relative_path)
        except Exception as e:
            if e.args[0] == 403 and e.args[1].get("errors")[0].get("code") == "too_large":
                return True
            return False
        return True
    else:
        return False


def _check_if_collab():
    try:
        if "fileId" in notebook_name():
            return True
    except Exception:
        return False


@dataclass
class GitVersion:
    repositoryName: str
    branchName: str
    commitHash: str
    commitComment: str
    commitAuthorName: str
    commitAuthorEmail: str
    isDirty: bool
    uri: str
    entrypoint: Optional[str] = None

    @classmethod
    def create(cls, path: str = ".", search_parent_directories: bool = True) -> Optional[GitVersion]:
        code_artifact = _extract_git_version(path, search_parent_directories)
        if code_artifact is None:
            return None
        return code_artifact

    @classmethod
    def create_from_github_uri(
        cls, uri: str, script_relative_path: Optional[str] = None, login_or_token=None, password=None, jwt=None
    ):
        return _create_git_version_from_uri(uri, script_relative_path, login_or_token, password, jwt)

    @classmethod
    def create_from_bitbucket_uri(
        cls,
        uri: str,
        script_relative_path: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        return _create_bitbucket_version_from_uri(uri, script_relative_path, username, password)

    @classmethod
    def create_from_gitlab_uri(
        cls, uri: str, script_relative_path: str, private_token: Optional[str] = None, oauth_token: Optional[str] = None
    ):
        return _create_gitlab_version_from_uri(uri, script_relative_path, private_token, oauth_token)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Tuple, Optional, TypeVar

from .metric import Metric


class WithMetricTrait(ABC):
    T = TypeVar("T", bound="WithMetricTrait")

    @abstractmethod
    def with_metric(self: T, key: str, value: float) -> T:
        """

        :param key:
        :param value:
        :return: iself
        """
        pass

    @abstractmethod
    def with_metrics(self: T, metrics: List[Tuple[str, float]]) -> T:
        """

        :param metrics:
        :return: iself
        """
        pass

    @abstractmethod
    def with_extended_metrics(self: T, metrics: List[Metric]) -> T:
        """

        :param metrics:
        :return: iself
        """
        pass


@dataclass
class WithMetrics(WithMetricTrait):
    metrics: Optional[List[Metric]] = None

    T = TypeVar("T", bound="WithMetrics")

    def with_metric(self: T, key: str, value: float) -> T:
        if key is not None and value is not None:
            if self.metrics is None:
                self.metrics = []
            self.metrics.append(Metric(key, value))
        return self

    def with_metrics(self: T, metrics: List[Tuple[str, float]]) -> T:
        if self.metrics is None:
            self.metrics = []
        for (key, value) in metrics:
            self.metrics.append(Metric(key, value))
        return self

    def with_extended_metrics(self: T, metrics: List[Metric]) -> T:
        if self.metrics is None:
            self.metrics = []
        if metrics is not None and len(metrics) > 0:
            self.metrics.extend(metrics)
        return self


class WithDelegatedMetrics(WithMetricTrait, ABC):
    T = TypeVar("T", bound="WithDelegatedMetrics")

    @abstractmethod
    def _get_delegate(self) -> WithMetricTrait:
        pass

    def with_metric(self: T, key: str, value: float) -> T:
        self._get_delegate().with_metric(key, value)
        return self

    def with_metrics(self: T, metrics: List[Tuple[str, float]]) -> T:
        self._get_delegate().with_metrics(metrics)
        return self

    def with_extended_metrics(self: T, metrics: List[Metric]) -> T:
        self._get_delegate().with_extended_metrics(metrics)
        return self



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class Dataset:
    name: str
    description: Optional[str] = None

    def with_description(self, description: str):
        """ """
        self.description = description
        return self



from dataclasses import dataclass, InitVar
from datetime import datetime
from typing import Optional

from vectice.models import dataset_version
from .dataset_output import DatasetOutput


@dataclass
class __Output:
    createdDate: datetime
    updatedDate: datetime
    id: int
    versionNumber: int
    name: str
    description: Optional[str]
    uri: Optional[str]
    authorId: int
    deletedDate: Optional[datetime]
    versionType: str
    dataSetId: int
    versionFolderId: int
    originId: Optional[int]
    isStarred: str
    dataSet: InitVar[dict]

    def __post_init__(self, dataSet: dict):
        self.dataSet = DatasetOutput(**dataSet)


@dataclass
class DatasetVersionOutput(dataset_version.DatasetVersion, __Output):
    pass



from dataclasses import dataclass, InitVar
from datetime import datetime
from typing import Optional

from vectice.models import JobRun
from .job_output import JobOutput


@dataclass
class __Output:
    createdDate: datetime
    updatedDate: datetime
    version: int
    id: int
    description: Optional[str]
    authorId: int
    duration: int
    metadataSource: str
    systemName: Optional[str]
    jobId: int
    deletedDate: Optional[datetime]
    job: InitVar[dict]
    name: Optional[str]

    def __post_init__(self, job: dict):
        self.job = JobOutput(**job)


@dataclass
class JobRunOutput(JobRun, __Output):
    pass



from typing import Optional

from vectice.entity import Artifact

from .job_run import RunApi
from .json_object import JsonObject


class ArtifactApi(RunApi):
    def __init__(
        self, project_token: str, job_id: int, run_id: int, _token: Optional[str] = None, allow_self_certificate=True
    ):
        super().__init__(
            project_token=project_token, job_id=job_id, _token=_token, allow_self_certificate=allow_self_certificate
        )
        self._run_id = run_id
        self._artifact_path = super().api_base_path + "/" + str(run_id) + "/artifact"

    @property
    def run_id(self) -> int:
        return self._run_id

    @property
    def api_base_path(self) -> str:
        return self._artifact_path

    def create_artifact(self, artifact: JsonObject) -> Artifact:
        if artifact.get("artifactType") is None:
            raise ValueError('"artifactType" must be provided in artifact.')
        if artifact.get("jobArtifactType") is None:
            raise ValueError('"jobArtifactType" must be provided in artifact.')
        return Artifact(self._post(self.api_base_path, artifact))

    def update_artifact(self, artifact_id: int, artifact: JsonObject) -> Artifact:
        return Artifact(self._put(self.api_base_path + "/" + str(artifact_id), artifact))



from typing import Optional
from urllib.parse import urlencode

from vectice.entity.dataset import Dataset

from .Page import Page
from .json_object import JsonObject
from .output.dataset_output import DatasetOutput
from .output.paged_response import PagedResponse
from .project import ProjectApi


class DatasetApi(ProjectApi):
    def __init__(self, project_token: str, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._dataset_path = super().api_base_path + "/dataset"

    @property
    def api_base_path(self) -> str:
        return self._dataset_path

    def list_datasets(
        self, search: str = None, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[DatasetOutput]:
        queries = {"index": page_index, "size": page_size}
        if search:
            queries["search"] = search
        datasets = self._get(self.api_base_path + "?" + urlencode(queries))
        return PagedResponse(
            item_cls=DatasetOutput, total=datasets["total"], page=datasets["page"], items=datasets["items"]
        )

    def create_dataset(self, dataset: JsonObject) -> Dataset:
        if dataset.get("name") is None:
            raise ValueError('"name" must be provided in dataset.')
        return Dataset(self._post(self.api_base_path, dataset))

    def update_dataset(self, dataset_id: int, dataset: JsonObject) -> Dataset:
        return Dataset(self._put(self.api_base_path + "/" + str(dataset_id), dataset))



from enum import Enum

from vectice.entity._versioned_entity import VersionedEntity


class Artifact(VersionedEntity):
    @property
    def job_artifact_type(self) -> str:
        return str(self["jobArtifactType"])

    @property
    def artifact_type(self) -> str:
        return str(self["artifactType"])

    @property
    def job_run_id(self) -> int:
        return int(self["jobRunId"])

    @property
    def job_property_id(self) -> int:
        return int(self["jobPropertyId"])

    @property
    def code_version_id(self) -> int:
        return int(self["codeVersionId"])

    @property
    def dataset_version_id(self) -> int:
        return int(self["dataSetVersionId"])

    @property
    def model_version_id(self) -> int:
        return int(self["modelVersionId"])

    @property
    def metadata_source(self) -> str:
        return str(self["metadataSource"])


class JobArtifactType(Enum):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"


class ArtifactType(Enum):
    PROPERTY = "PROPERTY"
    DATASET = "DATASET"
    MODEL = "MODEL"
    CODE = "CODE"



from vectice.entity._versioned_entity import VersionedEntity


class UserDeclaredVersion(VersionedEntity):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def version_number(self) -> int:
        return int(self["versionNumber"])

    @property
    def user_version(self) -> int:
        return int(self["userVersion"])

    @property
    def uri(self) -> int:
        return int(self["uri"])

    @property
    def is_favorite(self) -> bool:
        return bool(self["isFavourite"])



from typing import Optional, List, Tuple, Any, BinaryIO
from requests import Response

from vectice.api._auth import Auth
from .output.attachments_output import AttachmentOutput
from .output.paged_response import PagedResponse


class AttachmentApi(Auth):
    def __init__(
        self, _token: Optional[str] = None, auto_connect=True, allow_self_certificate=True, project_token=None
    ):
        super().__init__(_token=_token, auto_connect=auto_connect, allow_self_certificate=allow_self_certificate)
        self._attachment_path = "/metadata/entityfiles/"

    @property
    def api_base_path(self) -> str:
        return self._attachment_path

    def get_attachment(self, artifact_type: str, artifact_version: int, file_id: int):
        return self._get_attachment(self._attachment_path + f"{artifact_type}/{artifact_version}/{file_id}")

    def post_attachment(
        self, artifact_type: str, artifact_version: int, files: Optional[List[Tuple[str, Tuple[Any, BinaryIO]]]] = None
    ) -> Optional[Response]:
        if files and len(files) == 1:
            return self._post_attachments(self._attachment_path + f"{artifact_type}/{artifact_version}/", files)
        elif files and len(files) > 1:
            for file in files:
                self._post_attachments(self._attachment_path + f"{artifact_type}/{artifact_version}/", [file])
        return None

    def delete_attachment(self, artifact_type: str, artifact_version: int, file_id: int):
        return self._delete_attachment(self._attachment_path + f"{artifact_type}/{artifact_version}/{file_id}")

    def list_attachments(self, artifact_type: str, artifact_version: int) -> PagedResponse[AttachmentOutput]:
        attachments = self._list_attachments(self._attachment_path + f"{artifact_type}/{artifact_version}/")
        return PagedResponse(
            item_cls=AttachmentOutput,
            total=len(attachments),
            page={},
            items=attachments,
        )

    def update_attachments(
        self, artifact_type: str, artifact_version: int, files: List[Tuple[str, Tuple[Any, BinaryIO]]]
    ):
        attachments = {
            attach.fileName: attach.fileId for attach in self.list_attachments(artifact_type, artifact_version).list
        }
        for file in files:
            file_name = file[1][0]
            file_id = attachments.get(file_name)
            if file_id:
                self._update_attachments(
                    self._attachment_path + f"{artifact_type}/{artifact_version}/{file_id}", [file]
                )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, TypeVar


@dataclass
class WithParent:
    parentName: Optional[str] = None
    parentId: Optional[int] = None

    T = TypeVar("T", bound="WithParent")

    def with_parent_id(self: T, parent_id: int) -> T:
        """

        :param parent_id:
        :return: iself
        """
        self.parentName = None
        self.parentId = parent_id
        return self

    def with_parent_name(self: T, parent_name: str) -> T:
        """

        :param parent_name:
        :return: iself
        """
        self.parentName = parent_name
        self.parentId = None
        return self



from dataclasses import dataclass

from .artifact import _Base
from .git_version import GitVersion


@dataclass
class __CodeData:
    gitVersion: GitVersion
    """
    git information structure extracted automatically by the SDK.
    """


@dataclass
class CodeVersion(_Base, __CodeData):
    def __post_init__(self):
        self.parentName = self.gitVersion.repositoryName



from typing import Optional
from urllib.parse import urlencode

from vectice.api._auth import Auth

from .Page import Page
from .output.connection_output import ConnectionOutput
from .output.paged_response import PagedResponse


class ConnectionApi(Auth):
    def __init__(self, _token: Optional[str] = None, auto_connect=True, allow_self_certificate=True):
        super().__init__(_token=_token, auto_connect=auto_connect, allow_self_certificate=allow_self_certificate)
        self._connection_path = "/metadata/connection/"

    @property
    def api_base_path(self) -> str:
        return self._connection_path

    def _filter_response(self, response: dict):
        copy = response.copy()
        for idx, connection in enumerate(response["items"]):
            current_connection = connection.copy()
            for key, value in current_connection.items():
                if key not in ["name", "id", "type", "status"]:
                    copy["items"][idx].pop(key)
        return copy

    def list_connections(
        self,
        search_name: Optional[str] = None,
        connection_type: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[ConnectionOutput]:
        queries = {"index": page_index, "size": page_size}
        if search_name:
            queries["search"] = search_name
        if connection_type:
            queries["type"] = connection_type
        try:
            response = self._get(self._connection_path + "?" + urlencode(queries))
            connections = self._filter_response(response)
        except Exception as e:
            raise ConnectionError(f"Failed to retrieve connections due to {e}")
        return PagedResponse(
            item_cls=ConnectionOutput,
            total=connections["total"],
            page=connections["page"],
            items=connections["items"],
        )



from __future__ import annotations

from dataclasses import dataclass
from vectice.entity.model import ModelType
from typing import Optional


@dataclass
class Model:
    name: str
    type: ModelType
    description: Optional[str] = None

    def with_description(self, description: str):
        """"""
        self.description = description
        return self



import mimetypes
from enum import EnumMeta
from typing import Optional

mimetypes.init()
mimetypes.add_type("application/x-ipynb+json", ".ipynb")
mimetypes.add_type("text/csv", ".csv")


class DataResourcePathType(EnumMeta):
    Folder = "Folder"
    CsvFile = "CsvFile"
    ImageFile = "ImageFile"
    ExcelFile = "ExcelFile"
    TextFile = "TextFile"
    MdFile = "MdFile"
    DataSet = "DataSet"
    DataTable = "DataTable"
    File = "File"
    Notebook = "Notebook"

    @classmethod
    def get_file_type(cls, path: str):
        extension: Optional[str] = None
        try:
            extension = path.split(".", 1)[1].lower()
        except Exception:
            extension = None
        if extension == "csv":
            return DataResourcePathType.CsvFile
        # should we check stream with imghdr ?
        elif extension in ["jpg", "jpeg", "png", "webp", "bmp", "xbm", "tiff", "ppm", "pgm", "pbm", "gif", "exr"]:
            return DataResourcePathType.ImageFile
        elif extension == "md":
            return DataResourcePathType.MdFile
        elif extension == "ipynb":
            return DataResourcePathType.Notebook
        elif extension in ["txt", "text"]:
            return DataResourcePathType.TextFile
        elif extension in ["xls", "xlsx"]:
            return DataResourcePathType.ExcelFile
        else:
            return DataResourcePathType.File

    @classmethod
    def get_file_format(cls, path: str):
        return mimetypes.guess_type(path)[0]



from __future__ import annotations

import logging
from typing import Optional

from .artifact import Artifact, _Base
from .artifact_type import ArtifactType
from .code_version import CodeVersion
from .git_version import GitVersion


class CodeVersionArtifact(Artifact):
    def __init__(self, code: CodeVersion, description: Optional[str] = None):
        self.artifactType = ArtifactType.CODE
        self.description = description
        self.code: CodeVersion = code

    @classmethod
    def create(
        cls,
        path: str = ".",
    ) -> Optional[CodeVersionArtifact]:
        """
        create an artifact based on the git information relative to the given local path.

        :param path: the path to look for the git repository
        :return: a CodeVersion of None if a git repository was not found.
        """
        git_version = GitVersion.create(path)
        if git_version is not None:
            return cls(CodeVersion(git_version))
        elif not git_version:
            logging.warning(
                "Automatic code detection failed because the .git couldn't be found or file wasn't found by Github"
            )
            return None
        else:
            return None

    @classmethod
    def create_from_github_uri(
        cls, uri: str, script_relative_path: Optional[str] = None, login_or_token=None, password=None, jwt=None
    ) -> Optional[CodeVersionArtifact]:
        """
        create an artifact based on the github information relative to the given URI and relative path.

        Note: The URI given can include the branch you are working on. otherwise, the default repository branch will be used.

        sample :
            https://github.com/my-organization/my-repository (no branch given so using default branch)
            https://github.com/my-organization/my-repository/tree/my-current-branch (branch given is my-current-branch)

        To access private repositories, you need to authenticate with your credentials.
        see https://docs.github.com/en/github/authenticating-to-github/keeping-your-account-and-data-secure/about-authentication-to-github

        :param uri: the uri of the repository with a specific branch if needed.
        :param script_relative_path:  the file that is executed
        :param login_or_token: real login or personal access token
        :param password: the password
        :param jwt: the Oauth2 access token
        :return:
        """
        git_version = GitVersion.create_from_github_uri(uri, script_relative_path, login_or_token, password, jwt)
        if git_version is not None:
            return cls(CodeVersion(git_version))
        else:
            logging.warning(f"path {script_relative_path} is not part of a git repository")
            return None

    @classmethod
    def create_from_bitbucket_uri(
        cls,
        uri: str,
        script_relative_path: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> Optional[CodeVersionArtifact]:
        """
        Create a code artifact based on the Bitbucket information relative to the given URI and relative path.

        Note: The URI given can include the branch you are working on. otherwise, the default repository branch will be used.

        sample :
            https://bitbucket.org/workspace/project/ (no branch given so using default branch)
            https://bitbucket.org/workspace/project/src/branch (branch given is my-current-branch)

        To access private repositories, you need to authenticate with your credentials.
        see Bitbucket Cloud: https://atlassian-python-api.readthedocs.io/index.html

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param username: Bitbucket email
        :param password: Bitbucket password
        :return: A CodeVersion or None if the Bitbucket repository was not found or is not accessible
        """
        git_version = GitVersion.create_from_bitbucket_uri(uri, script_relative_path, username, password)
        if git_version is not None:
            return cls(CodeVersion(git_version))
        else:
            logging.warning(f"path {script_relative_path} is not part of a git repository")
            return None

    @classmethod
    def create_from_gitlab_uri(
        cls, uri: str, script_relative_path: str, private_token: Optional[str] = None, oauth_token: Optional[str] = None
    ) -> Optional[CodeVersionArtifact]:
        """
        Create an artifact that contains a version of a code

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param private_token: A real login or a personal access token
        :param oauth_token: The OAuth2 access token
        :return: A CodeVersion or None if the GitHub repository was not found or is not accessible
        """
        git_version = GitVersion.create_from_gitlab_uri(uri, script_relative_path, private_token, oauth_token)
        if git_version is not None:
            return cls(CodeVersion(git_version))
        else:
            logging.warning(f"path {script_relative_path} is not part of a git repository")
            return None

    def _get_delegate(self) -> _Base:
        return self.code



from __future__ import annotations

from dataclasses import dataclass
from typing import List, TypeVar, Optional, Tuple

from .tag import Tag


@dataclass
class WithTags:
    tags: Optional[List[Tag]] = None
    """
     list of tags associated with the artifact
    """

    T = TypeVar("T", bound="WithTags")

    def with_tag(self: T, key: str, value: str) -> T:
        """
        Add a tag to the artifact

        :param key: The key of the tag
        :param value: The value of the tag
        :return: itself
        """
        if self.tags is None:
            self.tags = []
        self.tags.append(Tag(key, value))
        return self

    def with_tags(self: T, tags: List[Tuple[str, str]]) -> T:
        """
        Add several tags to the artifact

        :param tags: A list of tuple `(key,value)`
        :return: itself
        """
        if self.tags is None:
            self.tags = []
        for (key, value) in tags:
            self.tags.append(Tag(key, value))
        return self



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from vectice.models import dataset


@dataclass
class __Output:
    id: int
    pattern: str
    isPatternBase: str
    createdDate: datetime
    updatedDate: datetime
    deletedDate: Optional[datetime]
    connectionId: int
    createdByUserId: int
    projectId: int
    workspaceId: int
    version: str


@dataclass
class DatasetOutput(dataset.Dataset, __Output):
    pass



from enum import EnumMeta


class ArtifactType(EnumMeta):
    """
    Enumeration of supported artifact types.
    """

    MODEL = "MODEL"
    """
    """

    CODE = "CODE"
    """
    """

    DATASET = "DATASET"
    """
    """



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Optional

from .job_run_status import JobRunStatus
from .with_properties import WithProperties
from .with_tags import WithTags


@dataclass
class JobRun(WithTags, WithProperties):
    startDate: datetime = field(default_factory=lambda: datetime.now(timezone.utc))
    """"""
    status: str = JobRunStatus.STARTED
    """"""
    endDate: Optional[datetime] = None
    """"""
    name: Optional[str] = None
    """"""



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Union

from .artifact import _Base
from .with_metrics import WithMetrics
from .with_properties import WithProperties
from .with_version import WithVersion
from ..entity.model import ModelType
from ..entity.model_version import ModelVersionStatus


@dataclass
class ModelVersion(_Base, WithVersion, WithProperties, WithMetrics):
    algorithmName: Optional[str] = None
    status: Optional[Union[ModelVersionStatus, str]] = None
    type: Optional[Union[ModelType, str]] = None
    """"""

    def with_algorithm(self, name: Optional[str]):
        """Accepts an optional str and sets the algorithmName.

        :param name: The model's algorithm
        """
        self.algorithmName = name

    def with_status(self, status: ModelVersionStatus):
        """Accepts an optional str and sets the model status.

        :param status: The model's status
        """
        self.status = status.name

    def with_type(self, model_type: ModelType):
        """Accepts an optional str and sets the model type.

        :param model_type: The model's type
        """
        self.type = model_type.name



from dataclasses import dataclass
from typing import Optional


@dataclass
class ArtifactVersion:
    versionNumber: Optional[int]
    """"""
    versionName: Optional[str]
    """"""
    id: Optional[int]
    """"""



from enum import EnumMeta


class JobType(EnumMeta):
    """
    Indicates the type of the job
    """

    EXTRACTION = "EXTRACTION"
    """
    """
    PREPARATION = "PREPARATION"
    """
    """
    TRAINING = "TRAINING"
    """
    """
    INFERENCE = "INFERENCE"
    """
    """
    DEPLOYMENT = "DEPLOYMENT"
    """
    """
    OTHER = "OTHER"
    """
    """



from __future__ import annotations

from typing import Optional, Union
from vectice.entity.model import ModelType
from vectice.entity.model_version import ModelVersionStatus

from .artifact import Artifact
from .artifact_type import ArtifactType
from .model_version import ModelVersion
from .with_metrics import WithDelegatedMetrics
from .with_properties import WithDelegatedProperties
from .with_version import WithDelegatedVersion
from .attachments import Attachments
from .with_attachments import WithDelegatedAttachments


class ModelVersionArtifact(
    Artifact[ModelVersion],
    WithDelegatedProperties,
    WithDelegatedVersion,
    WithDelegatedMetrics,
    WithDelegatedAttachments,
):
    def __init__(self, model: ModelVersion, attachments: Attachments, description: Optional[str] = None):
        self.artifactType = ArtifactType.MODEL
        self.description = description
        self.model: ModelVersion = model
        self.attachments: Attachments = attachments

    @classmethod
    def create(
        cls,
        description: Optional[str] = None,
    ) -> ModelVersionArtifact:
        """ """
        return cls(ModelVersion(), Attachments(cls.__name__), description)

    def _get_delegate(self) -> ModelVersion:
        return self.model

    def _get_attachment(self) -> Attachments:
        return self.attachments

    def with_algorithm(self, name: Optional[str]) -> ModelVersionArtifact:
        """Accepts an optional str and assigns it to the ModelVersion.

        :param name: The model's algorithm
        :return: ModelVersionArtifact
        """
        self._get_delegate().with_algorithm(name)
        return self

    def with_status(self, status: Union[ModelVersionStatus, str] = "EXPERIMENTATION") -> ModelVersionArtifact:
        """Accepts either an Enum or str, then checks if the
        option is valid. Then assigns the status to the ModelVersion.

        :param status: The model status
        :return: ModelVersionArtifact
        """
        try:
            model_status = status if isinstance(status, ModelVersionStatus) else ModelVersionStatus[status.upper()]
            self._get_delegate().with_status(model_status)
        except KeyError:
            raise ValueError(f"The status of {status} isn't supported.")
        return self

    def with_type(self, type: Union[ModelType, str] = "OTHER") -> ModelVersionArtifact:
        """Accepts either an Enum or str, then checks if the
        option is valid. Then assigns the type to the ModelVersion.

        :param type: The model type of its string representation
        :return: ModelVersionArtifact
        """
        try:
            model_type = type if isinstance(type, ModelType) else ModelType[type.upper()]
            self._get_delegate().with_type(model_type)
        except KeyError:
            raise ValueError(f"The type of {type} isn't supported.")
        return self

    def with_generated_version(self) -> ModelVersionArtifact:
        """Automatically increments the model version.

        :return: ModelVersionArtifact
        """
        self._get_delegate().with_generated_version()
        return self



from dataclasses import dataclass, InitVar, field
from typing import TypeVar, Generic, Sequence, Type, List

from vectice.api.Page import Page

ItemType = TypeVar("ItemType")


@dataclass
class PagedResponse(Generic[ItemType]):
    """
    Generic structure describing a result of a paged request.
    The structure contains page information and the list of items for this page.

    """

    total: int
    """
    total number of available pages
    """
    list: List[ItemType] = field(init=False)
    """
    current list of elements for this page
    """
    current_page: Page = field(init=False)
    """
    information on the current page.
    """
    page: InitVar[dict]
    item_cls: InitVar[Type[ItemType]]
    items: InitVar[Sequence[dict]]

    def __post_init__(self, page: dict, cls: Type[ItemType], items: Sequence[dict]):
        self.current_page = Page(**page)
        self.list = []
        if items is not None:
            for item in items:
                type_item = cls(**item)  # type: ignore
                self.list.append(type_item)
